"""# Page

The survey 'flow' is:

1. **Compile.** Execute a page's compile functions. By default, a page's 
first compile function runs its questions' compile methods in index order.

2. **Render.** Render the page for the participant and wait for them to 
respond.

3. **Record response.** Record the participant's response to every question on
the page. This sets the `response` attribute of the questions.

4. **Validate.** When the participant attempts to submit the page, validate 
the responses. As with the compile method, the validate method executes a 
page's validate functions in index order. By default, a page's first 
validate function runs its questions' validate methods in index order.

5. **Record data.** Record the data associated with the participant's
responses to every question on the page. This sets the `data` attribute of the
questions.

6. **Submit.** Execute the page's submit functions. By default, a page's 
first submit function runs its questions' submit methods in index order.

7. **Navigate.** If the page has a navigate function, create a new branch 
originating from this page.
"""

from ..app import db, settings
from ..tools import img
from .bases import BranchingBase
from .embedded import Timer
from .functions import Compile, Validate, Submit, Debug
from .private import CSSListType, JSListType
from .worker import Worker

from bs4 import BeautifulSoup
from flask import current_app, render_template, request
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import validates
from sqlalchemy_mutable import (
    HTMLAttrsType, MutableType, MutableListType, MutableListJSONType
)

import os
import webbrowser
from random import random, shuffle
from tempfile import NamedTemporaryFile
from time import sleep

DIR_PATH = os.path.dirname(os.path.realpath(__file__))

@Compile.register
def compile_questions(page):
    """
    Execute the page's questions' compile methods in index order.

    Parameters
    ----------
    page : hemlock.Page
    """
    [q._compile() for q in page.questions]

@Validate.register
def validate_questions(page):
    """
    Execute the page's questions' validate methods in index order.

    Parameters
    ----------
    page : hemlock.Page
    """
    [q._validate() for q in page.questions]
    
@Submit.register
def submit_questions(page):
    """
    Execute the page's questions' submit methods in index order.

    Parameters
    ----------
    page : hemlock.Page
    """
    [q._submit() for q in page.questions]

@Debug.register
def debug_questions(driver, page):
    """
    Execute the page's questions' debug methods in  *random* order.

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver

    page : hemlock.Page
    """
    order = list(range(len(page.questions)))
    shuffle(order)
    [page.questions[i]._debug(driver) for i in order]

@Debug.register
def navigate(driver, page, p_forward=.8, p_back=.1, sleep_time=3):
    """
    Randomly navigate forward or backward, or refresh the page. By default, it
    is executed after the default page debug function.

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver

    page : hemlock.Page

    p_forward : float, default=.8
        Probability of clicking the forward button.

    p_back : float, default=.1
        Probability of clicking the back button.

    sleep_time : float, default=3
        Number of seconds to sleep if there is no forward or back button on
        the page.

    Notes
    -----
    The probability of refreshing the page is `1-p_forward-p_back`.
    """
    forward_exists = back_exists = True
    if random() < p_forward:
        try:
            return driver.find_element_by_id('forward-btn').click()
        except:
            forward_exists = False
    if random() < p_back / (1 - p_forward):
        try:
            return driver.find_element_by_id('back-btn').click()
        except:
            back_exists = False
    if not (forward_exists or back_exists):
        sleep(sleep_time)
        navigate(driver, page, p_forward, p_back, sleep_time)
    driver.refresh()

settings['Page'] = dict(
    timer=None,
    template='hemlock/page.html',
    css=open(os.path.join(DIR_PATH, 'page-css.html'), 'r').read() \
        .splitlines(),
    js=open(os.path.join(DIR_PATH, 'page-js.html'), 'r').read() \
        .splitlines(),
    error_attrs={
        'class': ['alert', 'alert-danger', 'w-100', 'error'],
        'style': {'text-align': 'center'}
    },
    back=False,
    back_btn_attrs={
        'id': 'back-btn',
        'class': ['btn', 'btn-primary'],
        'name': 'direction',
        'type': 'submit',
        'style': {'float': 'left'},
        'value': 'back'
    },
    forward=True,
    forward_btn_attrs={
        'id': 'forward-btn',
        'class': ['btn', 'btn-primary'],
        'name': 'direction',
        'type': 'submit',
        'style': {'float': 'right'},
        'value': 'forward'
    },
    banner=img(
        '/hemlock/static/img/hemlock_banner.png',
        img_align='center', 
        img_attrs={'style': 'max-width:200px;', 'alt': 'Hemlock banner'}
    ),
    compile=[compile_questions],
    validate=[validate_questions],
    submit=[submit_questions],
    debug=[debug_questions, navigate],
    cache_compile=False,
    terminal=False,
)


class Page(BranchingBase, db.Model):
    """
    Pages are queued in a branch. A page contains a list of questions which it
    displays to the participant in index order.
    
    It inherits from [`hemlock.model.BranchingBase`](bases.md).

    Parameters
    ----------
    \*questions : list of hemlock.Question, default=[]
        Questions to be displayed on this page.

    template : str, default='hemlock/page-body.html'
        Template for the page `body`.

    extra_css : list, default=[]
        List of extra css elements to add to the default css.

    extra_js : list, default=[]
        List of extra javascript elements to add to the default 
        javascript.

    delay_forward : int or None, default=None
        Number of milliseconds for which to hide the forward button when the 
        page is first displayed.

    Attributes
    ----------
    back : str or None, default=None
        Text of the back button. If `None`, no back button will appear on the 
        page. You may also set `back` to `True`, which will set the text to 
        `'<<'`.

    back_btn_attrs : dict or None, default=None
        Dictionary of HTML attributes for the back button.

    banner : str or None
        Banner at the bottom of the page.

    cache_compile : bool, default=False
        Indicates that this page should cache the result of its compile 
        functions. Specifically, it removes all compile functions and the 
        compile worker from the page self `self._compile` is called.

    css : list
        List of css elements.

    direction_from : str or None, default=None
        Direction in which the participant navigated from this page. Possible 
        values are `'back'`, `'invalid'`, or `'forward'`.

    direction_to : str or None, default=None
        Direction in which the participant navigated to this page. Possible 
        values are `'back'`, `'invalid'`, and `'forward'`.

    error : str or None, default=None
        Error message.

    error_attrs : dict
        Dicitonary of HTML attributes for the error alert.

    forward : str or None, default='>>'
        Text of the forward button. If `None`, no forward button will appear 
        on the page. You may also set `forward` to `True`, which will set the 
        text to `'>>'`.

    forward_btn_attrs : dict
        Dictionary of HTML attributes for the forward button.

    g : misc, default=None
        A miscellaneous object.

    index : int or None, default=None
        Order in which this page appears in its branch's page queue.

    js : list
        List of javascript elements.

    navbar : str or None, default=None
        Navigation bar.

    terminal : bool, default=False
        Indicates that the survey terminates on this page.

    viewed : bool, default=False
        Indicates that the participant has viewed this page.

    Relationships
    -------------
    part : hemlock.Participant or None, default=None
        Participant to which this page belongs. Read only; derived from `self.
        branch`.

    branch : hemlock.Branch or None, default=None
        Branch to which this page belongs.

    next_branch : hemlock.Branch or None, default=None
        Branch which originated from this page. This is automatically set 
        when this page runs its navigate function.

    back_to : hemlock.Page or None
        Page to which this page navigates when going back. If `None`, this 
        page navigates back to the previous page.

    forward_to : hemlock.Page or None
        Page to which this page navigates when going forward. If `None`, this 
        page navigates to the next page.

    embedded : list of hemlock.Embedded, default=[]
        List of embedded data elements.

    timer : hemlock.Timer or None, default=hemlock.Timer
        Tracks timing data for this page.
        1. Timer can be set as a `Timer` object.
        2. Setting `timer` to a `str` or `None` sets the timer object 
        variable.
        3. Setting `timer` to a `tuple` sets the timer object variable and 
        data rows.
        4. Setting `timer` to a `dict` sets the timer object attributes. The
        `dict` maps attribute names to values.

    questions : list of hemlock.Question, default=[]
        List of questions which this page displays to its participant.

    data_elements : list of hemlock.DataElement
        List of data elements which belong to this page; in order, `self.
        timer`, `self.embedded`, `self.questions`.

    compile : list of hemlock.Compile
        List of compile functions; run before the page is rendered. The 
        default page compile function runs its questions' compile functions 
        in index order.

    compile_worker : hemlock.Worker or None, default=None
        Worker which sends the compile functions to a Redis queue.

    validate : list of hemlock.Validate
        List of validate functions; run to validate participant responses. 
        The default page validate function runs its questions' validate 
        functions in index order.

    validate_worker : hemlock.Worker or None, default=None
        Worker which sends the validate functions to a Redis queue.

    submit : list of hemlock.Submit
        List of submit functions; run after participant responses have been 
        validated. The default submit function runs its questions' submit 
        functions in index order.

    submit_worker : hemlock.Worker or None, default=None
        Worker which sends the submit functions to a Redis queue.

    navigate : hemlock.Navigate or None, default=None
        Navigate function which returns a new branch originating from this 
        page.

    navigate_worker : hemlock.Worker
        Worker which sends the navigate function to a Redis queue.

    debug_functions : list of hemlock.Debug
        List of debug functions; run during debugging. The default debug 
        function runs its questions' debug functions in *random* order.

    Examples
    --------
    ```python
    from hemlock import Page, push_app_context

    app = push_app_context()

    Page(Label('Hello World')).preview()
    ```

    Notes
    -----
    A CSS element can be any of the following:

    1. Link tag (str) e.g., `'<link rel="stylesheet" href="https://my-css-url">'`
    2. Href (str) e.g., `"https://my-css-url"`
    3. Style dictionary (dict) e.g., `{'body': {'background': 'coral'}}`

    The style dictionary maps a css selector to an attributes dictionary. The attributes dictionary maps attribute names to values.

    A javascript element can be any of the following:

    1. Attributes dictionary (dict) e.g., `{'src': 'https://my-js-url'}`
    2. JS code (str)
    3. Tuple of (attributes dictionary, js code)
    """
    id = db.Column(db.Integer, primary_key=True)
    
    # relationships
    @property
    def part(self):
        return None if self.branch is None else self.branch.part

    _branch_id = db.Column(db.Integer, db.ForeignKey('branch.id'))
    _branch_head_id = db.Column(db.Integer, db.ForeignKey('branch.id'))

    next_branch = db.relationship(
        'Branch',
        back_populates='origin_page',
        uselist=False,
        foreign_keys='Branch._origin_page_id'
    )
    
    _back_to_id = db.Column(db.Integer, db.ForeignKey('page.id'))
    back_to = db.relationship(
        'Page', foreign_keys=_back_to_id, remote_side=[id]
    )
        
    _forward_to_id = db.Column(db.Integer, db.ForeignKey('page.id'))
    forward_to = db.relationship(
        'Page', foreign_keys=_forward_to_id, remote_side=[id]
    )
    
    embedded = db.relationship(
        'Embedded',
        backref='page',
        order_by='Embedded.index',
        collection_class=ordering_list('index'),
    )

    _timer = db.relationship(
        'Timer', uselist=False, foreign_keys='Timer._timed_page_id'
    )

    @property
    def timer(self):
        return self._timer

    @timer.setter
    def timer(self, val):
        if isinstance(val, Timer):
            self._timer = val
            return
        if self._timer is None:
            self._timer = Timer()
        if isinstance(val, str) or val is None:
            self._timer.var = val
        elif isinstance(val, tuple):
            self._timer.var, self._timer.data_rows = val
        elif isinstance(val, dict):
            [setattr(self._timer, key, value) for key, value in val.items()]
        else:
            raise ValueError('Invalid value {} for timer'.format(val))        

    questions = db.relationship(
        'Question', 
        backref='page',
        order_by='Question.index',
        collection_class=ordering_list('index')
    )

    @property
    def data_elements(self):
        timer = [self.timer] if self.timer else []
        return self.embedded + timer + self.questions

    compile_worker = db.relationship(
        'Worker', uselist=False, foreign_keys='Worker._compile_id'
    )
    validate_worker = db.relationship(
        'Worker', uselist=False, foreign_keys='Worker._validate_id'
    )
    submit_worker = db.relationship(
        'Worker', uselist=False, foreign_keys='Worker._submit_id'
    )
    navigate_worker = db.relationship(
        'Worker', uselist=False, foreign_keys='Worker._navigate_page_id'
    )

    @validates(
        'compile_worker',
        'hello_world_worker', 
        'validate_worker', 
        'submit_worker', 
        'navigate_worker'
    )
    def _set_worker(self, key, worker):
        if not worker:
            return
        return worker if isinstance(worker, Worker) else Worker()

    # HTML attibutes
    template = db.Column(db.String)
    css = db.Column(CSSListType)
    js = db.Column(JSListType)
    navbar = db.Column(db.Text)
    error = db.Column(db.Text)
    error_attrs = db.Column(HTMLAttrsType)
    forward = db.Column(db.String)
    forward_btn_attrs = db.Column(HTMLAttrsType)
    back = db.Column(db.String)
    back_btn_attrs = db.Column(HTMLAttrsType)
    banner = db.Column(db.Text)

    @validates('forward', 'back')
    def _validate_direction(self, key, val):
        if not val:
            return None
        if val is True:
            return '&gt;&gt;' if key == 'forward' else '&lt;&lt;'
        return val

    # Function columns
    compile = db.Column(MutableListType)
    debug = db.Column(MutableListType)
    validate = db.Column(MutableListType)
    submit = db.Column(MutableListType)

    # Additional attributes
    cache_compile = db.Column(db.Boolean)
    run_compile = db.Column(db.Boolean, default=True)
    direction_from = db.Column(db.String(8))
    direction_to = db.Column(db.String(8))
    g = db.Column(MutableType)
    index = db.Column(db.Integer)
    name = db.Column(db.String)
    terminal = db.Column(db.Boolean)
    viewed = db.Column(db.Boolean, default=False)

    def __init__(
            self, *questions, extra_css=[], extra_js=[], delay_forward=None, 
            **kwargs
        ):
        def add_extra(attr, extra):
            # add extra css or javascript
            assert isinstance(extra, (str, list))
            if extra:
                assert isinstance(extra, (str, list))
                attr += [extra] if isinstance(extra, str) else extra

        self.questions = list(questions)
        self.run_compile = True
        super().__init__(**kwargs)
        add_extra(self.css, extra_css)
        add_extra(self.js, extra_js)
        if delay_forward:
            if self.forward_btn_attrs.get('style') is None:
                self.forward_btn_attrs['style'] = {}
            self.forward_btn_attrs['style']['display'] = 'none'
            self.js.append(
                render_template('hemlock/delay_forward.html', delay=delay_forward)
            )

    def clear_error(self):
        """
        Clear the error message from this page and all of its questions.

        Returns
        -------
        self : hemlock.Page
        """
        self.error = None
        [q.clear_error() for q in self.questions]
        return self
    
    def clear_response(self):
        """
        Clear the response from all of this page's questions.

        Returns
        -------
        self : hemlock.Page
        """
        [q.clear_response() for q in self.questions]
        return self
    
    def first_page(self):
        """
        Returns
        -------
        is_first_page : bool
            Indicator that this is the first page in its participant's survey.
        """
        if self.branch is None or self.index > 0:
            return False
        for branch in self.part.branch_stack:
            if branch.pages:
                first_nonempty_branch = branch
                break
        return self.branch == first_nonempty_branch

    def last_page(self):
        """
        Returns
        -------
        is_last_page : bool
            Indicator that this is the last page in its participant's survey.

        Notes
        -----
        This method assumes that if this page or its branch have a navigate
        function or next branch that this page is not the last (i.e. that the
        next branch will have pages). Avoid relying on this method if e.g. 
        this page's navigate function may return an empty branch.
        """
        if (
            self.branch is None 
            # not last page on the branch
            or self.index < len(self.branch.pages) - 1
            or self.navigate is not None
            or self.next_branch is not None
            or self.branch.navigate is not None
            or self.branch.next_branch is not None
        ):
            return False
        for branch in self.part.branch_stack:
            if self.branch == branch:
                return True
            if (
                branch.current_page is not None 
                and branch.current_page.last_page()
            ):
                return False

    def is_valid(self):
        """
        Returns
        -------
        valid : bool
            Indicator that all of the participant's responses are valid. That 
            is, that there are no error messages on the page or any of its 
            questions.
        """
        return not (self.error or any([q.error for q in self.questions]))

    def preview(self, driver=None):
        """
        Preview the page in a browser window.

        Parameters
        ----------
        driver : selenium.webdriver.chrome.webdriver.WebDriver or None, default=None
            Driver to preview page debugging. If `None`, the page will be
            opened in a web browser.

        Returns
        -------
        self : hemlock.Page

        Notes
        -----
        If running in WSL, first specify the distribution as an environment
        variable. For example, if running in Ubuntu:

        ```bash
        $ export WSL_DISTRIBUTION=Ubuntu
        ```

        This method does not run the compile functions.
        """
        def get_static_paths():
            # get a list of (static_url_path, static_folder tuples)
            paths = [(current_app.static_url_path, current_app.static_folder)]
            paths += [
                (bp.static_url_path, bp.static_folder)
                for bp in current_app.blueprints.values()
            ]
            return [path for path in paths if path[0]]

        def convert_rel_paths(attr):
            # convert relative paths to absolute paths
            elements = html.select('[{0}]:not([{0}=""])'.format(attr))
            [convert_rel_path(e, attr) for e in elements]

        def convert_rel_path(element, attr):
            # convert relative path for a single tag (element)
            url = element.attrs.get(attr)
            for path, folder in static_paths:
                if url.startswith(path):
                    element[attr] = dist + folder + url[len(path):]
                    break

        def create_tmp_file():
            # create tempoarary file and write out HTML
            with NamedTemporaryFile('w', suffix='.html', delete=False) as f:
                f.write(str(html))
                uri = 'file://'+('wsl$'+ dist + f.name if dist else f.name)
                if hasattr(current_app, 'tmpfiles'):
                    current_app.tmpfiles.append(f.name)
            return uri

        dist = os.environ.get('WSL_DISTRIBUTION', '')
        if not dist.startswith('/'):
            dist = '/'+dist
        static_paths = get_static_paths()
        html = BeautifulSoup(self._render(), 'lxml')
        convert_rel_paths('href')
        convert_rel_paths('src')
        uri = create_tmp_file()
        if driver is None:
            webbrowser.open(uri)
        else:
            driver.get(uri)
        return self

    def view_nav(self, indent=0):
        """
        Print the navigation starting at this page for debugging purposes.
        
        Parameters
        ----------
        indent : int, default=0
            Starting indentation.

        Returns
        -------
        self : hemlock.Page
        """
        # Note to self: the commented lines were very useful for me when 
        # debugging the navigation system; less so for users

        # HEAD_PART = '<== head page of participant'
        # HEAD_BRANCH = '<== head page of branch'
        HEAD_PART = 'C '
        head_part = HEAD_PART if self == self.part.current_page else ''
        # head_branch = HEAD_BRANCH if self == self.branch.current_page else ''
        # print(' '*indent, self, head_branch, head_part)
        terminal = 'T' if self.terminal else ''
        print(' '*indent, self, head_part+terminal)
        if self.next_branch in self.part.branch_stack:
            self.next_branch.view_nav()
        return self

    # methods executed during study
    def _compile(self):
        """
        Run the page's compile functions. If `self.cache_compile`, the page's 
        compile functions and compile worker are removed.

        Returns
        -------
        self
        """
        if self.run_compile:
            # the page uses its compile functions to f itself
            [f(self) for f in self.compile]
            if self.cache_compile:
                self.compile = self.compile_worker = None
            self.run_compile = False
        return self
    
    def _render(self):
        return render_template(self.template, page=self)

    def _record_response(self):
        """Record participant response
        
        Begin by updating the total time. Then get the direction the 
        participant requested for navigation (forward or back). Finally, 
        record the participant's response to each question.
        """
        self.direction_from = request.form.get('direction')
        [q._record_response() for q in self.questions]
        return self
    
    def _validate(self):
        """Validate response
        
        Check validate functions one at a time. If any returns an error 
        message (i.e. error is not None), indicate the response was invalid 
        and return False. Otherwise, return True.
        """
        self.clear_error()
        for f in self.validate:
            self.error = f(self)
            if self.error:
                break
        is_valid = self.is_valid()
        self.direction_from = 'forward' if is_valid else 'invalid'
        return is_valid
    
    def _submit(self):
        [q._record_data() for q in self.questions]
        [f(self) for f in self.submit]
        return self

    def _debug(self, driver):
        [f(driver, self) for f in self.debug]
        return self


"""Login"""

from ...app import bp, db
from ...models import Page
from ...qpolymorphs import Input

from flask import current_app, redirect, request, session, url_for
from werkzeug.security import check_password_hash

from functools import wraps

@bp.route('/login', methods=['GET','POST'])
def login():
    """Login view function"""
    def login_page():
        error = None
        if request.args.get('incorrect_password'):
            error = 'The password you entered was incorrect'
        elif request.args.get('requested'):
            error = 'Login required to access this page.'
        return Page(
            Input(
                '<p>Please enter your password</p>', 
                type='password', key='password'
            ),
            error=error, forward='Login'
        )._render()

    if request.method == 'GET':
        return login_page()
    # request method is POST
    requested = request.args.get('requested', 'status')
    session['password'] = request.form.get('password')
    if password_is_correct():
        return redirect(url_for('hemlock.{}'.format(requested)))
    return redirect(
        url_for('hemlock.login', requested=requested, incorrect_password=True)
    )

def password_is_correct():
    """Indicate that the session password is correct"""
    if not current_app.config.get('PASSWORD'):
        return not session.get('password')
    if 'password' not in session:
        return False
    return check_password_hash(
        current_app.config.get('PASSWORD_HASH'), session['password']
    )

def researcher_login_required(func):
    """Decorator requiring researcher login"""
    @wraps(func)
    def login_required():
        if password_is_correct():
            return func()
        return redirect(url_for('hemlock.login', requested=func.__name__))

    return login_required

@bp.route('/logout')
def logout():
    if 'download-page-id' in session:
        page = Page.query.get(session['download-page-id'])
        if page:
            db.session.delete(page)
            db.session.commit()
    session.clear()
    return redirect(url_for('hemlock.login'))


"""# Branch"""

from ..app import db
from .bases import BranchingBase
from .worker import Worker

from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import validates
from sqlalchemy_mutable import MutableType


class Branch(BranchingBase, db.Model):
    """
    Branches are stacked in a participant's branch stack. A branch contains a 
    queue of pages which it displays to its participant.

    Inherits from [`hemlock.models.Base`](bases.md).

    Parameters
    ----------
    \*pages : hemlock.Page
        Pages which belong to this branch.

    Attributes
    ----------
    index : int or None, default=None
        Order in which this branch appears in its participant's branch stack.

    Relationships
    -------------
    part : hemlock.Participant
        Participant to whose branch stack this page belongs.

    origin_branch : hemlock.Branch
        The branch from which this branch originated.

    next_branch : hemlock.Branch
        The branch which originated from this branch.

    origin_page : hemlock.Page
        The page from which this branch originated. Note that branches can originate from other branches or pages.

    pages : list of hemlock.Page
        The queue of pages belonging to this branch.

    start_page : hemlock.Page or None
        The first page in the page queue, if non-empty.

    current_page : hemlock.Page
        Current page of this branch (head of the page queue).

    embedded : list of hemlock.Embedded
        Embedded data elements.

    data_elements : list of hemlock.DataElement
        All data elements belonging to this branch, in order of embedded data 
        then page data.

    navigate : hemlock.Navigate
        Navigate function which returns a new branch once the participant has 
        reached the end of this branch (i.e. the end of the page queue 
        associated with this branch).

    navigate_worker : hemlock.Worker
        Worker which handles complex navigate functions.

    Examples
    --------
    ```python
    from hemlock import Branch, Label, Page, push_app_context

    app = push_app_context()

    Branch(
    \    Page(Label('Hello World')),
    \    Page(Label('Hello Moon')),
    \    Page(Label('Hello Star'))
    ).preview()
    ```

    This will open all of the branch's pages in separate tabs.
    """
    id = db.Column(db.Integer, primary_key=True)
    
    _part_id = db.Column(db.Integer, db.ForeignKey('participant.id'))
    _part_head_id = db.Column(db.Integer, db.ForeignKey('participant.id'))
    
    _origin_branch_id = db.Column(db.Integer, db.ForeignKey('branch.id'))
    origin_branch = db.relationship(
        'Branch',
        back_populates='next_branch',
        uselist=False,
        foreign_keys='Branch._origin_branch_id'
    )
        
    _next_branch_id = db.Column(db.Integer, db.ForeignKey('branch.id'))
    next_branch = db.relationship(
        'Branch',
        back_populates='origin_branch',
        uselist=False,
        remote_side=id,
        foreign_keys='Branch._next_branch_id'
    )
    
    _origin_page_id = db.Column(db.Integer, db.ForeignKey('page.id'))
    origin_page = db.relationship(
        'Page', 
        back_populates='next_branch',
        foreign_keys='Branch._origin_page_id'
    )
    
    pages = db.relationship(
        'Page', 
        backref='branch', 
        order_by='Page.index',
        collection_class=ordering_list('index'),
        foreign_keys='Page._branch_id'
    )
    
    @property
    def start_page(self):
        """Return the start of the page queue"""
        return self.pages[0] if self.pages else None
        
    current_page = db.relationship(
        'Page', 
        uselist=False,
        foreign_keys='Page._branch_head_id'
    )
        
    embedded = db.relationship(
        'Embedded', 
        backref='branch',
        order_by='Embedded.index',
        collection_class=ordering_list('index')
    )
        
    @property
    def data_elements(self):
        elements = []
        elements += self.embedded
        [elements.extend(p.data_elements) for p in self.pages]
        return elements
        
    navigate_worker = db.relationship(
        'Worker', uselist=False, foreign_keys='Worker._navigate_branch_id'
    )

    @validates('navigate_worker')
    def _set_worker(self, key, worker):
        if not worker:
            return
        return worker if isinstance(worker, Worker) else Worker()

    index = db.Column(db.Integer)

    def __init__(self, *pages, **kwargs):
        self.pages = list(pages)
        super().__init__(**kwargs)

    def preview(self, driver=None):
        """
        Preview the page queue in the a browser window.

        Parameters
        ----------
        driver : selenium.webdriver.chrome.webdriver.WebDriver or None, default=None
            Driver to preview page debugging. If `None`, the page will be
            opened in a web browser.

        Returns
        -------
        self : helock.Branch
        """
        [p.preview(driver) for p in self.pages]
        self

    def view_nav(self):
        """
        Print this branch's page queue for debugging purposes.
        
        Returns
        -------
        self : hemlock.Branch
        """
        # Note to self: the commented lines were very useful for me when 
        # debugging the navigation system; less so for users

        # HEAD_PART = '<== head branch of participant'
        # HEAD_BRANCH = '<== head page of branch'
        indent = 4*(0 if self.index is None else self.index)
        # head_part = HEAD_PART if self == self.part.current_branch else ''
        # print(' '*indent, self, head_part)
        print(' '*indent, self)
        [p.view_nav(indent) for p in self.pages]
        # head_branch = HEAD_BRANCH if None == self.current_page else ''
        # print(' '*indent, None, head_branch)
        if self.next_branch in self.part.branch_stack:
            self.next_branch.view_nav()
        return self
        
    def _forward(self):
        """Advance forward to the next page in the queue"""
        if self.current_page is not None:
            new_head_index = self.current_page.index + 1
            if new_head_index == len(self.pages):
                self.current_page = None
            else:
                self.current_page = self.pages[new_head_index]
        return self
    
    def _back(self):
        """Return to previous page in queue"""
        if self.pages:
            if self.current_page is None:
                self.current_page = self.pages[-1]
            else:
                new_head_index = self.current_page.index - 1
                self.current_page = self.pages[new_head_index]
        return self


"""# Randomization tools"""

from flask_login import current_user

import numpy as np

import random
import time
from itertools import combinations, cycle, permutations, product
from operator import itemgetter
from random import choice, shuffle
from string import ascii_letters, digits

def key(len_=90):
    """
    Parameters
    ----------
    len_ : int, default=90
        Length of the key to generate.

    Returns
    -------
    key : str
        Randomly generated key of ascii letters and digits of specificed 
        length.

    Notes
    -----
    The first character is a letter. This allows you to use `key` to generate
    strongly random id's for html elements. (Html element id's cannot start
    with a digit.)

    Examples
    --------
    ```python
    from hemlock import tools

    tools.key(10)
    ```

    Out:

    ```
    gpGmZuRfF7
    ```
    """
    chars = ascii_letters + digits
    first_char  = choice(ascii_letters)
    return first_char + ''.join(choice(chars) for _ in range(len_-1))

def reset_random_seed():
    seed = (1000 * int(time.time())) % 2**32
    random.seed(seed)
    np.random.seed(seed)

class Randomizer():
    """
    Evenly randomizes over a set of elements.

    Parameters
    ----------
    elements : iterable
        Set of elements over which to randomize.

    r : int, default=1
        Size of the subset of elements to select.

    combination : bool, default=True
        Indicates randomization over combinations of the elements, as opposed 
        to permutations.

    Attributes
    ----------
    elements : iterable
        Set from the `elements` parameter.

    Examples
    --------
    ```python
    from hemlock.tools import Randomizer

    elements = ('world','moon','star')
    randomizer = Randomizer(elements, r=2, combination=False)
    randomizer.next()
    ```

    Out:

    ```
    ('moon', 'world')
    ```
    """
    def __init__(self, elements, r=1, combination=True):
        self.elements = elements
        idx = list(range(len(elements)))
        iter_ = combinations(idx, r) if combination else permutations(idx, r)
        iter_ = list(iter_)
        shuffle(iter_)
        self._iter = cycle(iter_)
        
    def next(self):
        """
        Returns
        -------
        subset :
            Selected subset of elements.
        """
        return itemgetter(*next(self._iter))(self.elements)


class Assigner(Randomizer):
    """
    Evenly assigns participants to conditions. Inherits from 
    `hemlock.tools.Randomizer`.

    Parameters
    ----------
    conditions : dict
        Maps condition variable name to iterable of possible assignments.

    Attributes
    ----------
    keys : iterable
        Condition variable names.

    elements : iterable
        All possible combinations of condition values to which a participant may be assigned.

    Examples
    --------
    ```python
    from hemlock import Participant, push_app_context
    from hemlock.tools import Assigner

    push_app_context()

    part = Participant.gen_test_participant()
    conditions = {'Treatment': (0,1), 'Level': ('low','med','high')}
    assigner = Assigner(conditions)
    assigner.next()
    ```

    Out:

    ```
    {'Treatment': 1, 'Level': 'low'}
    ```

    In:

    ```python
    [(e.var, e.data) for e in part.embedded]
    ```

    Out:

    ```
    [('Treatment', 0), ('Level', 'low')]
    ```
    """

    def __init__(self, conditions):
        self.keys = conditions.keys()
        super().__init__(tuple(product(*conditions.values())))
        
    def next(self, participant=None):
        """
        Assigns the participant to a condition. The condition assigment 
        updates the participant's metadata.

        Parameters
        ----------
        participant : hemlock.Participant or None, default=None
            This method records the assignment in this participant's `meta`
            dictionary. If `None`, the participant is the current user.

        Returns
        -------
        assignment : dict
            Maps condition variable names to assigned conditions.

        Notes
        -----
        By default, this method assigns the participant using flask-login's
        `current_user`. If you're assigning the participant in function 
        handled by Redis, this won't work. You'll need to manually pass in the
        participant.
        """        
        assignment = super().next()
        assignment = {key: val for key, val in zip(self.keys, assignment)}
        try:
            participant = participant or current_user
            participant.meta.update(assignment)
        except:
            print('Unable to update participant metadata.')
        return assignment


"""# Debugger

The debugger sends 'AI participants' through the survey. The AI 
participants attempt to break the survey by clicking random objects and 
entering random responses.

Examples
--------
This example debugs an app locally.

In `survey.py`:

```python
from hemlock import Branch, Page, Label, route

@route('/survey')
def start():
\    x = 1/0
\    return Branch(Page(Label('Hello World'), terminal=True))
```

In `app.py`:

```python
import survey

from hemlock import create_app

app = create_app()

if __name__ == '__main__':
\    from hemlock.app import socketio
\    socketio.run(app)
```

Open a terminal and run the app with:

```
$ python app.py $ or python3 app.py
```

Open a second terminal and open the python shell with:

```
$ python # or python3
```

Run the debugger in the second terminal:

```
>>> from hemlock.debug import AIParticipant, debug
>>> debug()
```

The debugger will open a chromedriver and attempt to complete the survey. The
first terminal window will display this error:

```
File "/home/<username>/hemlock/survey.py", line 9, in start
\    x = 1/0
ZeroDivisionError: division by zero
```

Notes
-----
If your app is running on a different local host port than 5000, set the url 
root as an environment variable before opening the python shell in your second
terminal:

```
$ export ULR_ROOT=http://localhost:xxxx
```

If your application import is not `app.app`, set the import as an enviornment
variable before opening the python shell in your second terminal:

```
$ export APP_IMPORT=path.to.app
```

AI participants run in batches of specified sizes. For local debugging, I 
recommend a batch size of 1. For production debugging, you can safely go up 
to 3.
"""

from ..models import Participant, Page
from ..tools import chromedriver

import os
import sys
import unittest
import warnings
from pydoc import locate
from threading import Thread
from time import sleep

SERVER_ERR = 'Internal Server Error'
WERKZEUG_ERR = "Brought to you by DON'T PANIC, your friendly Werkzeug powered traceback interpreter."
ERROR_MSG = "{ai.part} encountered an error. \n\nDon't give up."

def debug(num_batches=1, batch_size=1):
    """
    Run the debugger.

    Parameters
    ----------
    num_batches : int, default=1
        Number of batches of AI participants to run.

    batch_size : int, default=1
        Number of AI participants to run per batch.

    Returns
    -------
    result : bool
        `True` if all AI participants in all batches run sucessfully.
        Otherwise, the program will crash.

    Notes
    -----
    When called from the command line tool, `num_batches` and `batch_size` are
    passed as strings.
    """
    results = [run_batch(int(batch_size)) for i in range(int(num_batches))]
    return all(results)

def run_batch(batch_size=1):
    """
    Run a batch of AI participants.
    
    Parameters
    ----------
    batch_size : int, default=1
        Number of AI participants to run in this batch.

    Returns
    -------
    result : bool
        `True` if all AI participants in this batch run successfully. 
        Otherwise, the program will crash.
    """
    threads = [
        SuccessThread(target=run_participant, daemon=True)
        for i in range(batch_size)
    ]
    [t.start() for t in threads]
    [t.join() for t in threads]
    if not all([t.success for t in threads]):
        sys.exit()
    return True

def run_participant():
    """
    Run a single AI participant through the survey. Assert that the
    participant does not encounter failures or errors.

    Returns
    -------
    result : bool
        `True` if the participant ran successfully.
    """
    result = unittest.main(exit=False).result
    assert not result.failures and not result.errors
    return True


class AIParticipant(unittest.TestCase):
    def setUp(self):
        """Push application context and connect webdriver"""
        warnings.simplefilter('ignore', ResourceWarning)
        sys.path.insert(0, os.getcwd())
        app = locate(os.environ.get('APP_IMPORT') or 'app.app')
        sys.path.pop(0)
        self.ctx = app.app_context()
        self.ctx.push()
        self.driver = chromedriver()
        url_root = os.environ.get('URL_ROOT') or 'http://localhost:5000'
        self.driver.get(url_root+'?Test=1')
        self.part = Participant.query.all()[-1]
        super().setUp()

    def test(self):
        """An AI Participant has encountered an error."""
        current_page = self.get_current_page()
        while current_page is None or not current_page.terminal:
            self.check_for_error()
            if current_page is None:
                self.driver.refresh()
                sleep(3)
            else:
                print('Debugging page ', current_page, current_page.name)
                current_page._debug(self.driver)
            current_page = self.get_current_page()

    def get_current_page(self):
        """Get the current page"""
        try:
            while True:
                # accept all alerts at the start of the page
                self.driver.switch_to.alert.accept()
        except:
            pass
        try:
            page_tag = self.driver.find_element_by_tag_name('page')
        except:
            return None
        page_id = page_tag.get_attribute('id').split('-')[-1]
        return Page.query.get(page_id)

    def check_for_error(self):
        """Assert that there are no errors on this page"""
        error = False
        try:
            h1 = self.driver.find_element_by_tag_name('h1')
            error = h1.text == SERVER_ERR
        except:
            pass
        try:
            footer = self.driver.find_element_by_css_selector('div.footer')
            error = error or footer.text == WERKZEUG_ERR
        except:
            pass
        assert not error, ERROR_MSG.format(ai=self)

    def tearDown(self):
        """Close webdriver and pop application context"""
        self.driver.close()
        self.ctx.pop()
        super().tearDown()


class SuccessThread(Thread):
    """Operates as normal thread with success indicator"""
    def run(self):
        try:
            super().run()
            self.success = True
        except:
            self.success = False


"""# Statics

Tool for embedding statics:

- css
- javascript
- images
- iframes
- YouTube videos
- links to Google bucket files
"""

from flask import current_app

import re
import os
from datetime import timedelta
from urllib.parse import parse_qs, urlparse, urlencode

DIR = os.path.dirname(os.path.realpath(__file__))

def format_attrs(**attrs):
    """
    Parameters
    ----------
    \*\*attrs : dict
        Maps HTML attribute name to value.

    Returns
    -------
    attrs : str
        Formated attributes for insertion into HTML tag.
    """
    def format_item(key, val):
        if not val:
            # val is empty string, None, False, etc.
            return ''
        return key if val is True else '{}="{}"'.format(key, val)

    attrs = attrs.copy()
    if 'class' in attrs and isinstance(attrs['class'], list):
        attrs['class'] = ' '.join(attrs['class'])
    if 'style' in attrs and isinstance(attrs['style'], dict):
        attrs['style'] = ' '.join(
            ['{}:{};'.format(*item) for item in attrs['style'].items()
        ])
    return ' '.join(format_item(*item) for item in attrs.items())

def external_css(href, rel='stylesheet', type='text/css', **attrs):
    """
    Parameters
    ----------
    \*\*attrs :
        Attribute names and values in the `<link/>` tag.

    Returns
    -------
    css : str
        `<link/>` tag.

    Examples
    --------
    ```python
    from hemlock import Page, push_app_context
    from hemlock.tools import external_css

    app = push_app_context()

    p = Page(extra_css=external_css(href='https://my-css-url'))
    p.css
    ```

    Out:

    ```
    ...
    <link href="https://my-css-url" rel="stylesheet" type="text/css"/>
    ```
    """
    return '<link {}/>'.format(
        format_attrs(href=href, rel=rel, type=type, **attrs)
    )

def internal_css(style):
    """
    Parameters
    ----------
    style : dict
        Maps css selector to an attributes dictionary. The attributes 
        dictionary maps attribute names to values.

    Returns
    -------
    css : str
        `<style>` tag.

    Examples
    --------
    ```python
    from hemlock import Page, push_app_context
    from hemlock.tools import internal_css

    app = push_app_context()

    p = Page(extra_css=internal_css({'body': {'background': 'coral'}}))
    p.css
    ```

    Out:

    ```
    ...
    <style>
    \    body {background:coral;}
    </style>
    ```
    """
    def format_style(selector, attrs):
        attrs = ' '.join(['{}: {};'.format(*item) for item in attrs.items()])
        return selector+' {'+attrs+'}'

    css = ' '.join([format_style(key, val) for key, val in style.items()])
    return '<style>{}</style>'.format(css)

def external_js(src, **attrs):
    """
    Parameters
    ----------
    src : str
        External javascript source.

    \*\*attrs :
        Attribute names and values in the `<script>` tag.

    Returns
    -------
    js : str
        `<script>` tag.

    Examples
    --------
    ```python
    from hemlock import Page, push_app_context
    from hemlock.tools import external_js

    app = push_app_context()

    p = Page(extra_js=external_js(src='https://my-js-url'))
    p.js
    ```

    Out:

    ```
    ...
    <script src="https://my-js-url"></script>
    ```
    """
    return '<script {}></script>'.format(format_attrs(src=src, **attrs))

def internal_js(js, **attrs):
    """
    Parameters
    ----------
    js : str
        Javascript code.

    \*\*attrs : dict
        Mapping of HTML attributes to values for the `<script>` tag.

    Returns
    -------
    js : str
        Javascript code wrapped in `<script>` tag.

    Examples
    --------
    ```python
    from hemlock import Page, push_app_context
    from hemlock.tools import internal_js

    app = push_app_context()

    p = Page(
    \    extra_js=internal_js(
    \        '''
    \        $( document ).ready(function() {
    \            alert('hello, world!');
    \        });
    \        '''
    \    )
    )
    p.js
    ```

    Out:

    ```
    ...
    <script>
    \    $( document ).ready(function() {
    \        alert('hello, world!');
    \    });
    </script>
    ```
    """
    if not js.startswith('<script>'):
        js = '<script {}>{}</script>'.format(format_attrs(**attrs), js)
    return js

NO_BUCKET_ERR_MSG = """
Enable a Google Cloud Bucket to access this feature.
\n  See `hlk gcloud-bucket`
"""

def src_from_bucket(filename):
    """
    Parameters
    ----------
    filename : str
        Name of the file in the Google bucket.

    Returns
    -------
    src : str
        `src` html attribute which references the specified file in the Google
        bucket.

    Examples
    --------
    Set up a 
    [Google bucket](https://cloud.google.com/storage/docs/creating-buckets)
    with the appropriate 
    [CORS permissions](https://cloud.google.com/storage/docs/cross-origin).

    Set an environment variable `BUCKET` to the name of the bucket.

    ```
    $ export BUCKET=<my-bucket>
    ```

    Upload a file to the bucket, e.g. <https://xkcd.com/2138/> and name it 
    `wanna_see_the_code.png`.

    ```python
    from hemlock import Branch, Page, Label, push_app_context
    from hemlock.tools import Img, src_from_bucket

    app = push_app_context()

    img = Img(
    \    src=src_from_bucket('wanna_see_the_code.png'),
    \    align='center'
    ).render()
    Page(Label(img)).preview()
    ```
    """
    bucket = os.environ.get('BUCKET')
    assert bucket is not None, NO_BUCKET_ERR_MSG
    return '/'.join(['https://storage.cloud.google.com', bucket, filename])

def url_from_bucket(filename, expiration=1800, **kwargs):
    """
    Parameters
    ----------
    filename : str
        Name of the file in the Google bucket.

    expiration : float, default=1800
        Number of seconds until the url expires.

    \*\*kwargs :
        Keyword arguments are passed to the [`generate_signed_url` method]
        (https://cloud.google.com/storage/docs/access-control/signed-urls).

    Returns
    -------
    signed_url : str
        Signed url for the file in the app's bucket.

    Examples
    --------
    Set up a 
    [Google bucket](https://cloud.google.com/storage/docs/creating-buckets)
    with the appropriate 
    [CORS permissions](https://cloud.google.com/storage/docs/cross-origin).

    Set an environment variable `BUCKET` to the name of the bucket, and 
    `GOOGLE_APPLICATION_CREDENTIALS` to the name of your 
    [Google application credentials JSON file](https://cloud.google.com/docs/authentication/getting-started).

    ```
    $ export BUCKET=<my-bucket> GOOGLE_APPLICATION_CREDENTIALS=<my-credentials.json>
    ```

    In `survey.py`:

    ```python
    from hemlock import Branch, Page, Download, route
    from hemlock.tools import url_from_bucket

    @route('/survey')
    def start():
    \    filename = 'wanna_see_the_code.png'
    \    url = url_from_bucket(filename)
    \    return Branch(Page(Download(downloads=[(url, filename)])))
    ```
    
    In `app.py`:

    ```python
    import survey

    from hemlock import create_app

    app = create_app()

    if __name__ == '__main__':
    \    from hemlock.app import socketio
    \    socketio.run(app, debug=True)
    ```

    Run the app locally with:

    ```
    $ python app.py # or python3 app.py
    ```

    And open your browser to <http://localhost:5000/>. Click on the 
    download button to download the file from your Google bucket.
    """
    assert hasattr(current_app, 'gcp_bucket'), NO_BUCKET_ERR_MSG
    kwargs['expiration'] = timedelta(expiration)
    blob = current_app.gcp_bucket.blob(filename)
    return blob.generate_signed_url(**kwargs)

def img(
    src, caption='', img_align='left', caption_align='left',
    figure_class='figure w-100', figure_attrs={},
    img_class='figure-img img-fluid rounded', img_attrs={},
    caption_class='figure-caption', caption_attrs={},
    template=os.path.join(DIR, 'img.html')
):
    """
    Parameters
    ----------
    src : str
        Image source.

    caption : str

    img_align : str, default='left'

    caption_align : str, default='left'

    figure_class : list or str, default='figure w-100'
        Class for the `<figure>` tag.

    figure_attrs : dict, default={}
        HTML attributes for the `<figure>` tag.

    img_class : list or str, default='figure-img img-fluid rounded'
        Class for the `<img>` tag.

    img_attrs : dict, default={}
        HTML attributes for the `<img>` tag.

    caption_class : list or str, default='figure-caption'
        Class for the `<figcaption>` tag.

    caption_attrs : dict, default={}
        HTML attributes for the `<figcaption>` tag.

    template : str, default='directory/img.html'
        Path to image template. Default is a template in the same directory as
        the current file. This may also be a string to be used directly as the
        template.

    Returns
    -------
    img : str
        Image html.

    Examples
    --------
    ```python
    from hemlock.tools import img

    img(
    \    'https://imgs.xkcd.com/comics/wanna_see_the_code.png',
    \    img_align='center', caption='Wanna See The Code?'
    )
    ```

    Out:

    ```
    <figure class="figure w-100 text-center">
    \    <img class="figure-img img-fluid rounded" src="https://imgs.xkcd.com/comics/wanna_see_the_code.png">
    \    <figcaption class="figure-caption text-left">Wanna See The Code?</figcaption>
    </figure>
    ```
    """
    _add_class(figure_attrs, figure_class)
    figure_attrs['class'].append('text-'+img_align)
    _add_class(img_attrs, img_class)
    img_attrs['src'] = src
    _add_class(caption_attrs, caption_class)
    caption_attrs['class'].append('text-'+caption_align)
    if os.path.exists(template):
        template = open(template).read()
    return template.format(
        figure_attrs=format_attrs(**figure_attrs),
        img_attrs=format_attrs(**img_attrs),
        caption_attrs=format_attrs(**caption_attrs),
        caption=caption
    )

def iframe(
    src, aspect_ratio=(16, 9), query_string={},
    div_class='embed-responsive', div_attrs={},
    iframe_class='embed-responsive-item', 
    iframe_attrs={'allowfullscreen': True},
    template=os.path.join(DIR, 'iframe.html')
):
    """
    Parameters
    ----------
    src : str
        Embed source.

    aspect_ratio : tuple of (int, int), default=(16, 9)
        Embed 
        [aspect ratio](https://getbootstrap.com/docs/4.0/utilities/embed/#aspect-ratios).

    query_string : dict
        Mapping of URL query keys to values.

    div_class : str or list, default='embed-responsive'
        Class of the `<div>` which wraps the `<iframe>`.

    div_attrs : dict, default={}
        HTML attributes for the `<div>` wrapper.

    iframe_class : str or list, default='embed-responsive-item'
        Class of the `<iframe>` tag.

    iframe_attrs : dict, default={'allowfullscreen': True}
        HTML attributes for the `<iframe>` tag.

    template : str, default='directory/iframe.html'
        Path to iframe template. Default is a template in the same directory 
        as the current file. This may also be a string to be used directly as 
        the template.

    Returns
    -------
    iframe : str
        Iframe HTML.

    Examples
    --------
    ```python
    from hemlock.tools import iframe

    iframe(
    \    'https://www.youtube.com/embed/zpOULjyy-n8?rel=0', 
    \    aspect_ratio=(21, 9)
    )
    ```

    Out:

    ```
    <div class="embed-responsive embed-responsive-21by9">
    \    <iframe allowfullscreen class="embed-responsive-item" src="https://www.youtube.com/embed/zpOULjyy-n8?rel=0"></iframe>
    </div>
    ```
    """
    _add_class(div_attrs, div_class)
    div_attrs['class'].append('embed-responsive-{}by{}'.format(*aspect_ratio))
    _add_class(iframe_attrs, iframe_class)
    if query_string:
        src += '?' + urlencode(query_string)
    iframe_attrs['src'] = src
    if os.path.exists(template):
        template = open(template).read()
    return template.format(
        div_attrs=format_attrs(**div_attrs),
        iframe_attrs=format_attrs(**iframe_attrs)
    )

def youtube(
    src, iframe_attrs={
        'allow': 'accelerometer; autoplay; encrypted-media; gyroscope; picture-in-picture;', 
        'allowfullscreen': True
    }, *args, **kwargs
):
    """
    Embed a youtube video.

    Parameters
    ----------
    src : str
        Link to youtube video. e.g. 
        <https://www.youtube.com/watch?v=ts3s738ZkcQ>.

    iframe_attrs : dict
        HTML attributes for the `<iframe>` tag.

    \*args, \*\*kwarg : 
        Arguments and keyword arguments are passed to `hemlock.tools.iframe`.

    Returns
    -------
    iframe : str
        Iframe HTML.

    Examples
    --------
    ```python
    from hemlock.tools import youtube

    youtube(
    \    'https://www.youtube.com/watch?v=ts3s738ZkcQ', 
    \    query_string={'autoplay': 1}
    )
    ```

    Out:

    ```
    <div class="embed-responsive embed-responsive-16by9">
    \    <iframe allow="accelerometer; autoplay; encrypted-media; gyroscope; picture-in-picture;" allowfullscreen class="embed-responsive-item" src="https://www.youtube.com/embed/ts3s738ZkcQ?autoplay=1"></iframe>
    </div>
    ```
    """
    video_id = parse_qs(urlparse(src).query).pop('v')[0]
    src = 'https://www.youtube.com/embed/{}'.format(video_id)
    return iframe(src, *args, iframe_attrs=iframe_attrs, **kwargs)

def _add_class(attrs, class_):
    attrs['class'] = class_ if isinstance(class_, list) else [class_]


"""Tools"""

from .comprehension import comprehension_check
from .lang import indef_article, join, plural, pronouns
from .mturk import (
    consent_page, completion_page, get_approve_df, approve_assignments
)
from .navbar import Navbar, Navitem, Navitemdd, Dropdownitem
from .random import Assigner, Randomizer, key, reset_random_seed
from .statics import (
    format_attrs, external_css, internal_css, external_js, internal_js, 
    src_from_bucket, url_from_bucket, img, iframe, youtube
)
from .titrate import titrate
from .utils import (
    chromedriver, get_data, html_list, html_table, show_on_event, url_for
)


"""Data Profile"""

from ...app import bp
from ...models import Page, Worker
from ...models.private import DataStore
from .login import researcher_login_required
from .utils import navbar, render, researcher_page

import pandas as pd
import pandas_profiling
from flask import current_app

EMPTY_DATAFRAME = "No data are available to create the profle."

PROFILE_CREATION_ERR = '''
An error occurred while creating the data profile.<br/>
This may be because too few participants have completed the survey, or because you have not enabled background jobs.
'''

@bp.route('/profile')
@researcher_login_required
def profile():
    profile_p = profile_page()
    if not DataStore.query.first().data:
        profile_p.error = EMPTY_DATAFRAME
        return profile_p._render()
    return render_profile(profile_p)

@researcher_page('profile')
def profile_page():
    return Worker.compile(Page(
        compile_functions=create_profile, back=False, forward=False, g={}
    ))

def render_profile(profile_p):
    """Render the data profile

    Return the worker's loading page while the profile is being created. If 
    there is a error, return an error message. Otherwise, return the 
    rendered profile.

    Note that an error will occur if the researcher has not provisioned a 
    Redis server.
    """
    worker = profile_p.compile_worker
    try:
        if not worker.job_finished:
            return worker()
        worker.reset()
    except:
        profile_p.error = PROFILE_CREATION_ERR
        return profile_p._render()
    profile_report = profile_p.g.get('profile_report')
    if profile_report is None:
        profile_p.error = PROFILE_CREATION_ERR
        return profile_p._render()
    return profile_report.to_html()

def create_profile(profile_p):
    """Create the data profile"""
    df = pd.DataFrame(DataStore.query.first().data)
    if current_app.settings.get('clean_data'):
        df = current_app.clean_data(df)
    try:
        profile_p.g = {'profile_report': df.profile_report()}
    except:
        pass


"""# Submit functions"""

from ..models import Submit
from .utils import (
    convert, correct_choices as correct_choices_, match as match_
)

@Submit.register
def correct_choices(question, *values):
    """
    Convert the question's data to a 0-1 indicator that the participant 
    selected the correct choice(s).

    Parameters
    ----------
    question : hemlock.ChoiceQuestion

    \*values :
        Values of the correct choices.

    Notes
    -----
    If the participant can only select one choice, indicate whether the 
    participant selected one of the correct choices.

    Examples
    --------
    ```python
    from hemlock import Check, Submit as S, push_app_context

    app = push_app_context()

    check = Check(
    \    '<p>Select the correct choice.</p>',
    \    ['correct', 'incorrect', 'also incorrect'],
    \    submit=S.correct_choices('correct')
    )
    check.response = check.choices[0]
    check._submit().data
    ```

    Out:

    ```
    1
    ```
    """
    question.data = int(correct_choices_(question, *values))

@Submit.register
def data_type(question, new_type, *args, **kwargs):
    """
    Convert the quesiton's data to a new type. If the question's data cannot
    be converted, it is changed to `None`.
    
    Parameters
    ----------
    question : hemlock.Question

    new_type : class
    
    \*args, \*\*kwargs :
        Arguments and keyword arguments to pass to the `new_type` constructor.

    Examples
    --------
    ```python
    from hemlock import Input, Submit as S, push_app_context

    app = push_app_context()

    inpt = Input(data='1', submit=S.data_type(int))
    inpt._submit()
    inpt.data, isinstance(inpt.data, int)
    ```

    Out:

    ```
    (1, True)
    ```
    """
    question.data, success = convert(question.data, new_type, *args, **kwargs)
    if not success:
        question.data = None

@Submit.register
def match(question, pattern):
    """
    Convert the question's data to a 0-1 indicator that the data matches the
    pattern.

    Parameters
    ----------
    question : hemlock.Question

    pattern : str or hemlock.Question
        Regex pattern to match. If this is a `Question`, the pattern is the 
        question's `response`'.

    Examples
    --------
    ```python
    from hemlock import Input, Submit as S, push_app_context

    app = push_app_context()

    inpt = Input(data='hello world', submit=S.match('hello *'))
    inpt._submit().data
    ```

    Out:

    ```
    1
    ```
    """
    question.data = int(match_(question, pattern) is not None)


"""Question polymorphs"""

from .blank import Blank
from .check import Binary, Check
from .choice import Choice, Option
from .dashboard import Dashboard
from .download import Download
from .file import File
from .input import Input
from .label import Label
from .range import Range, RangeInput
from .select import Select
from .textarea import Textarea


"""Participant status"""

from ...app import bp
from ...models import Page
from ...models.private import DataStore
from ...qpolymorphs import Label
from ...tools import external_js
from .login import researcher_login_required
from .utils import navbar

from flask import current_app, url_for

PARTICIPANTS = """
<table align="center" style="width:50%">
    <tr>
        <th style="text-align:center" colspan="2">Participants' Current Status</th>
    </tr>
    <tr><td colspan="2"><hr/></td></tr>
    <tr>
        <td>Completed</td>
        <td id="Completed" align="right">{Completed}</td>
    </tr>
    <tr>
        <td>In Progress</td>
        <td id="InProgress" align="right">{InProgress}</td>
    </tr>
    <tr>
        <td>Timed Out</td>
        <td id="TimedOut" align="right">{TimedOut}</td>
    </tr>
    <tr>
        <td>Total</td>
        <td id="Total" align="right">{Total}</td>
    </tr>
</table>
"""

@bp.route('/status')
@researcher_login_required
def status():
    """View participants' live status"""
    ds = DataStore.query.first()
    return Page(
        Label(PARTICIPANTS.format(**ds.current_status)),
        navbar=navbar.render(), back=False, forward=False,
        extra_js=[
            external_js(current_app.settings['socket_js_src']),
            external_js(url_for('hemlock.static', filename='js/status.js'))
        ]
    )._render()


"""Utilities"""

from selenium_tools import get_datetime, send_datetime

import math
from datetime import MAXYEAR, MINYEAR, datetime, timedelta
from random import choice, randint, random, randrange, shuffle
from string import ascii_letters, digits

import re

def convert(obj, type_, *args, **kwargs):
    """
    Converts an object to a new type.

    Parameters
    ----------
    obj :
        Object to convert.

    type_ : class or None
        Desired new type.

    \*args, \*\*kwargs :
        Arguments and keyword arguments to pass to the `type_` constructor.

    Returns
    -------
    obj :
        Converted object, if possible, else original object.

    success : bool
        Indicate that the conversion was successful.
    """
    if type_ is None:
        return obj, True
    try:
        converted = type_(obj, *args, **kwargs)
        return converted, True
    except:
        return obj, False

def correct_choices(q, *values):
    """
    Parameters
    ----------
    q : hemlock.ChoiceQuestion

    \*values : 
        Values of the correct choices.

    Returns
    -------
    correct : bool
        Indicates that the participant selected the correct choice(s).

    Notes
    -----
    If the participant can only select one choice, indicate whether the 
    participant selected one of the correct choices. If the participant
    selected no choices, indicate whether there were any correct values
    available to select.
    """
    if q.response is None or q.response == []:
        return not bool(values)
    data = (
        [c.value for c in q.response] if isinstance(q.response, list) 
        else q.response
    )
    return set(data) == set(values) if q.multiple else data in values

def match(question, pattern):
    """
    Check for a full regular expression match of the question response against 
    the pattern.

    Parameters
    ----------
    question : hemlock.Question
        Question whose response should match the pattern.

    pattern : str or hemlock.Question
        Pattern the response should match. If this is a `Question`, this function will look for a full match to the pattern question's `response`.

    Returns
    -------
    full match : bool
        Indicates that a full match was found.
    """
    from ..models import Question

    def resp_to_str(response):
        return str(response) if response is not None else ''

    if isinstance(pattern, Question):
        pattern = resp_to_str(pattern.response)
    return re.fullmatch(str(pattern), resp_to_str(question.response))

def random_datetime(inpt, min=datetime(1900, 1, 1), max=datetime(2100, 1, 1)):
    def get_dt(attr):
        val = inpt.get_attribute(attr)
        if val:
            return get_datetime(type_, val)

    type_ = inpt.get_attribute('type')
    start = get_dt('min') or min
    stop = get_dt('max') or max
    delta = int((stop - start).days * 24* 60 * 60) # resolution in seconds
    send_datetime(inpt, start + timedelta(seconds=randint(0, delta)))

def random_num(inpt, min=-1000, max=1000, step=.001, p_int=.5):
    start = inpt.get_attribute('min')
    start = min if start in (None, '') else float(start)
    stop = inpt.get_attribute('max')
    stop = max if stop in (None, '') else float(stop)
    step_ = inpt.get_attribute('step')
    step_ = step if step_ in (None, '') or step_ == 'any' else float(step_)
    x = start + random() * (stop - start)
    x = round(x / step_) * step_
    inpt.send_keys(str(int(x) if random() < p_int else x))

def random_str(inpt, maxlength=100, p_whitespace=.2):
    chars = ascii_letters + digits
    chars = list(chars) + [' '] * int(p_whitespace*len(chars))
    # response length follows exponential distribution
    maxlength_ = inpt.get_attribute('maxlength')
    maxlength_ = maxlength if not maxlength_ else float(maxlength_)
    magnitude = int(math.log(maxlength_))
    length = int(random() * 10**randint(1,magnitude))
    inpt.send_keys(''.join([choice(chars) for i in range(length)]))


"""# Worker"""

from ..app import db
from .bases import Base

from flask_worker import WorkerMixin


class Worker(WorkerMixin, Base, db.Model):
    """
    Workers simplify interaction with a Redis queue. A worker attaches to a 
    branch or page, and is responsible for one of its methods (compile, 
    validate, submit, or navigate).

    When the method for which a worker is responsible is called, the worker
    sends the method to a Redis queue. While the Redis queue is processing 
    this method, the worker shows participants a loading page. When the Redis
    queue finishes processing this method, the worker sends the client to his
    next page.

    Worker inherits from 
    [`flask_worker.WorkerMixin`](https://dsbowen.github.io.flask-worker/).

    Examples
    --------
    In `survey.py`:

    ```python
    from hemlock import Branch, Compile as C, Label, Page, route

    @route('/survey')
    def start():
    \    return Branch(
    \        Page(
    \            Label('Hello, World!')
    \        ),
    \        Page(
    \            Label(
    \                'Goodbye, Moon!',
    \                compile=C.complex_function(seconds=5)
    \            ),
    \            compile_worker=True,
    \            terminal=True
    \        ),
    \    )

    @C.register
    def complex_function(label, seconds):
    \    import time
    \    for t in range(seconds):
    \        print('Progress: {}%'.format(round(100.*t/seconds)))
    \        time.sleep(1)
    \    print('Progress: 100%')
    ```

    Note that the second page (or rather, one of its questions), needs to run
    a complex compile function. We add a worker to it by setting 
    `compile_worker=True`. Use a similar syntax to add validate, submit, and 
    navigate workers.

    Our `app.py` is standard:

    ```python
    import survey

    from hemlock import create_app

    app = create_app()

    if __name__ == '__main__':
    \    from hemlock.app import socketio
    \    socketio.run(app, debug=True)
    ```

    To run the app locally, you will need to set the `REDIS_URL` environment 
    variable and run a redis queue from your project's root directory.

    **Note.** Windows cannot run redis natively. To run redis on Windows, use 
    [Windows Subsystem for Linux](../setup/wsl.md).

    If using the hemlock template and hemlock-CLI:

    1. Open `env.yaml` and add the line `REDIS_URL: redis://`.
    2. Open a second terminal in your project's root directory and run the redis queue with `hlk rq`.
    3. Run the app by entering `hlk serve` in your first terminal.

    If not using the template or hemlock-CLI:

    1. Set your environment variable with `export REDIS_URL=redis://`.
    2. Open a second terminal in your project's root directory and enter `rq worker hemlock-task-queue`.
    3. Run the app by entering `python3 app.py` in your first terminal.

    Go to <http://localhost:5000/> in your browser. Notice that, when you click past the first page, you see a loading gif before the second page is loaded. In your second terminal window, you should see:

    ```
    Progress: 0%
    ...
    Progress: 100%
    ```

    Notes
    -----

    To run redis in production on heroku:

    <b>1. Declare a worker process</b>
    
    In the root directory of your project, open `Procfile`. Add the line `worker: rq worker -u $REDIS_URL hemlock-task-queue`. Your `Procfile` should look like:

    ```
    web: gunicorn -k eventlet app:app
    worker: rq worker -u $REDIS_URL hemlock-task-queue
    ```

    <b>2. Create a redis addon and provision worker processes</b>

    In the root directory of your project, open `app.json`.

    To create a Redis addon, add `{"plan": "heroku-redis:hobby-dev", "options": {"version": 5}}` to your `addons`. To provision worker processes, add `"worker: {"quantity": 1, "size": "hobby"}` to your `formation`. In sum, the top of `app.json` will look like:

    ```json
    {
    \    "addons": [
    \        "heroku-postgresql:hobby-dev", 
    \        {"plan": "heroku-redis:hobby-dev", "options": {"version": 5}}
    \    ],
    \    "formation": {
    \        "web": {"quantity": 1, "size": "hobby"},
    \        "worker": {"quantity": 1, "size": "hobby"}
    \    },
        ...
    ```

    When scaling for production, I recommend using the premium-1 redis plan, 
    quantity 5, size standard-1x. So `app.json` will look like:

    ```json
    {
    \    "addons": [
    \        "heroku-postgresql:standard-1x", 
    \        {"plan": "heroku-redis:premium-1", "options": {"version": 5}}
    \    ],
    \    "formation": {
    \        "web": {"quantity": 10, "size": "standard-1x"},
    \        "worker": {"quantity": 5, "size": "standard-1x"}
    \    },
        ...
    ```

    Notes
    -----
    I recommend redis version 5, rather than the latest version 6. Why? Heroku required you to manually set up TLS authentication for redis 6, which is a complicated process, whereas you can bypass TLS authentication using redis 5.
    """
    id = db.Column(db.Integer, primary_key=True)
    _compile_id = db.Column(db.Integer, db.ForeignKey('page.id'))
    _validate_id = db.Column(db.Integer, db.ForeignKey('page.id'))
    _submit_id = db.Column(db.Integer, db.ForeignKey('page.id'))
    _navigate_page_id = db.Column(db.Integer, db.ForeignKey('page.id'))
    _navigate_branch_id = db.Column(db.Integer, db.ForeignKey('branch.id'))
    _router_id = db.Column(db.Integer, db.ForeignKey('router.id'))


"""# Utilities"""

from cssutils import parseStyle
from flask import render_template, url_for as try_url_for
from selenium.webdriver import Chrome
from selenium.webdriver.chrome.options import Options

import os

DIR = os.path.dirname(os.path.realpath(__file__))

def chromedriver(headless=False):
    """
    Parameters
    ----------
    headless : bool, default=False
        Indicates whether to run Chrome in headless mode.

    Returns
    -------
    driver : selenium.webdriver.chrome.webdriver.WebDriver

    Notes
    -----
    Chromedriver must be headless in production. When the application is in 
    production, this method sets `headless` to `True` regardless of the 
    parameter you pass.

    Examples
    --------
    ```python
    from hemlock.tools import chromedriver

    driver = chromedriver()
    ```
    """
    options = Options()
    url_root = os.environ.get('URL_ROOT')
    if headless or (url_root and 'localhost' not in url_root):
        options._arguments = ['--disable-gpu', '--no-sandbox', '--headless']
    else: 
        options._arguments = ['--disable-gpu', '--no-sandbox']
    options.binary_location = os.environ.get('GOOGLE_CHROME_BIN', '')
    driver_path = os.environ.get('CHROMEDRIVER_PATH')
    return (
        Chrome(driver_path, options=options) if driver_path 
        else Chrome(options=options)
    )

def get_data(dataframe='data'):
    """
    Parameters
    ----------
    dataframe : str, default='data'
        Name of the dataframe to get; `'data'` or `'meta'`.

    Returns
    -------
    data : dict
        Maps variable names to list of entries. May not include data from
        participants who are in progress.

    Examples
    --------
    ```python
    from hemlock import Branch, Page, Participant, push_app_context
    from hemlock.tools import get_data

    def start():
    \    return Branch(Page())

    push_app_context()

    Participant.gen_test_participant(start).completed = True
    get_data()
    ```

    Out:

    ```
    {'ID': [1],
    'EndTime': [datetime.datetime(2020, 7, 6, 17, 11, 0, 245032)],
    'StartTime': [datetime.datetime(2020, 7, 6, 17, 11, 0, 245032)],
    'Status': ['Completed']}
    ```
    """
    from ..models.private import DataStore
    return dict(getattr(DataStore.query.first(), dataframe))

def html_list(*items, ordered=True):
    template = '<ol>\n{}\n</ol>' if ordered else '<ul>\n{}\n</ul>'
    return template.format(
        '\n'.join(['<li>{}</li>'.format(item) for item in items])
    )

def html_table(table, extra_classes=[]):
    """
    Parameters
    ----------
    table : dict
        Maps column names to a list of entries.

    Returns
    -------
    table : str
        Table in HTML format.

    Examples
    --------
    ```python
    from hemlock.tools import html_table

    html_table({
        'Column 0': ['hello', 'world'],
        'Column 1': ['goodbye', 'moon']
    })
    ```

    Out:

    ```
    <table class="table">
    \    <thead>
    \        <tr>
    \            <th scope="col">Column 0</th>
    \            <th scope="col">Column 1</th>
    \        </tr>
    \    </thead>
    \    <tbody>
    \        <tr>
    \            <td>hello</td>
    \            <td>goodbye</td></tr>
    \        <tr>
    \            <td>world</td>
    \            <td>moon</td></tr>
    \    </tbody>
    \</table>
    ```
    """
    def table_row(row):
        return tr_template.format(
            '\n'.join(
                [td_template.format(items[row]) for items in items_list]
            )
        )

    template = open(os.path.join(DIR, 'table.html')).read()
    tr_template = '<tr>{}</tr>'
    th_template = '<th scope="col">{}</th>'
    td_template = '<td>{}</td>'
    head = '\n'.join([th_template.format(col) for col in table.keys()])
    # assume all items lists are the same length
    items_list = list(table.values())
    body = '\n'.join([table_row(i) for i in range(len(items_list[0]))])
    classes = ' '.join(['table'] + extra_classes)
    return template.format(classes, head, body)

def show_on_event(
        target, condition, value, init_hidden=True, *args,  **kwargs
    ):
    """
    Show the target question when a condition is met.

    Parameters
    ----------
    target : hemlock.Question
        The question which will be shown when the condition is met.

    condition : hemlock.Question
        The question whose value determines whether the target is shown.

    value : str or hemlock.Choice
        1. If the condition is an input, the target is shown when the input 
        value matches this value.
        2. If the condition has choices, the target is shown when the choice
        with this value is checked.

    init_hidden : bool, defualt=True
        Indicates that the initial state of the target should be hidden.

    regex : bool, default=False
        Indicates that the target will be shown if input value matches the 
        string as a regular expression.

    event : str or None, default=None
        Type of event which toggles the target display. If `None`, this 
        function infers the type of event based on inputs.

    duration : str or int, default=400
        Show/hide event duration in milliseconds.

    Examples
    --------
    ```python
    from hemlock import Page, Check, Input, Label, push_app_context
    from hemlock.tools import show_on_event

    app = push_app_context()

    race = Check(
    \    '<p>Indicate your race.</p>', 
    \    ['White', 'Black', 'Other'], 
    \    multiple=True
    )
    specify = Input('<p>Please specify.</p>')
    show_on_event(specify, race, 'Other')
    Page(race, specify).preview()
    ```

    ```python
    name = Input("<p>What's your name?</p>")
    greet = Label("Hello, World!")
    show_on_event(greet, name, '(w|W)orld', regex=True, duration=400)
    Page(name, greet).preview()
    ```
    """
    from ..models import Question
    from ..qpolymorphs import Choice, Option

    # NOTE I want to extend this to show choices and options as well
    # but the code below isn't working right now

    # assert (
    #     isinstance(target, (Question, Choice, Option)), 
    #     'target must be a Question, Choice, or Option'
    # )
    # if isinstance(target, Question):
    #     # form-group
    #     target_id = target.model_id+'-fg'
    # elif isinstance(target, Choice):
    #     # custom-conctrol
    #     target_id = target.model_id+'-cc'
    # else:
    #     # option
    #     target_id = target.model_id
    assert isinstance(target, Question), 'target must be a Question'
    if init_hidden:
        if 'style' not in target.form_group_attrs:
            target.form_group_attrs['style'] = {}
        target.form_group_attrs['style']['display'] = 'none'
    target.js.append(
        _show_on_event_js(target, condition, value, *args, **kwargs)
    )
    return target

def _show_on_event_js(
        target, condition, value, regex=False, duration=400, event=None, 
    ):
    from ..models import ChoiceQuestion
    from ..qpolymorphs import Binary, Check, Select

    def get_event_value(condition, value):
        if isinstance(condition, ChoiceQuestion):
            event = 'change' # listen for change event
            for choice in condition.choices:
                if choice.value == value:
                    value = choice.key
                    break
        else:
            event = 'input' # listen for input event
        return event, value

    event, value = get_event_value(condition, value)
    choice = option = False # value is neither a radio, check box, or option
    if isinstance(condition, (Binary, Check)):
        # value corresponds to radio or check box
        choice, option = True, False
    elif isinstance(condition, Select):
        # value corresponds to option
        choice, option = False, True        
    return render_template(
        'hemlock/show-on-event.js', 
        target=target, 
        condition=condition,
        choice=choice,
        option=option,
        value=value, 
        regex=regex, 
        event=event, 
        duration=duration
    )

def url_for(*args, **kwargs):
    """
    Attempt to return `flask.url_for(*args, **kwargs)`. However, this method 
    does not exit the program when getting a url outside a request context; 
    e.g. when debugging in a shell or notebook.

    Parameters
    ----------
    \*args, \*\*kwargs :
        Arguments and keyword arguments will be passed to `flask.url_for`.

    Returns
    -------
    url : str
        Output of `flask.url_for` if possible; otherwise `'URL_UNAVAILABLE'`.
    """
    try:
        return try_url_for(*args, **kwargs)
    except:
        return 'URL_UNAVAILABLE'


"""Routes for participants"""

from ...app import bp, db, login_manager, socketio
from ...models import Participant, Page
from ...qpolymorphs import Label
from ...models.private import DataStore

from datetime import datetime, timedelta
from flask import current_app, redirect, render_template, request, url_for
from flask_login import current_user, login_required, login_user, logout_user

"""Initial views and functions"""
@login_manager.user_loader
def load_user(id):
    return Participant.query.get(id)

@bp.route('/')
def index():
    """
    Direct visitors to survey, restart page, or screenout page.

    Direct to screenout page if:

    1. Visitor metadata is in `app.screenouts`, or
    2. Participant has already completed the survey and is no longer logged in

    Direct to restart page if:

    1. Participant has started the survey and the restart option is available

    Otherwise, direct to default survey route.
    """
    meta = get_metadata()
    if is_screenout(meta):
        return redirect(url_for('hemlock.screenout'))
    
    in_progress = current_user.is_authenticated
    if in_progress and current_user.current_branch is not None:
        if current_user.current_page.first_page():
            # no need to restart if on first page
            return redirect(url_for('hemlock.'+bp.default_route))
        if not current_app.settings['restart_option']:
            return redirect(url_for('hemlock.'+bp.default_route))
        return redirect(url_for('hemlock.restart', **meta))
    if is_duplicate(meta):
        return redirect(url_for('hemlock.screenout'))

    initialize_participant(meta)
    return redirect(url_for('hemlock.'+bp.default_route))

def get_metadata():
    """
    This is where it gets meta.

    Returns
    -------
    meta : dict
        Maps URL parameter names to values. This will capture things like
        worker ID for MTurk studies. Additionally, capture IPv4.
    """
    meta = dict(request.args)
    ip = request.environ.get('HTTP_X_FORWARDED_FOR', None)
    meta['IPv4'] = request.remote_addr if ip is None else ip.split(',')[0]
    return meta

def initialize_participant(meta):
    """Initialize Participant with given metadata
    
    If there is a time limit, start the clock.
    """
    if current_user.is_authenticated:
        logout_user()
    part = Participant(meta=meta)
    db.session.add(part)
    db.session.commit()
    login_user(part)
    
    # FIX THIS
    # if current_app.time_limit is None:
    #     return
    # end_time = datetime.now() + current_app.time_limit
    # current_app.apscheduler.add_job(
    #     func=time_out, 
    #     trigger='date', 
    #     run_date=end_time, 
    #     args=(current_app._get_current_object(), part.id), 
    #     id=str(part.id)
    # )
        
def time_out(app, part_id):
    """Indicate Participant time has expired"""
    with app.app_context():
        part = Participant.query.get(part_id)
        part.time_expired = True
        db.session.commit()

"""Screenout and duplicate handling"""
def is_screenout(meta):
    """
    Returns
    -------
    match_found : bool
        Indicates that the visitor metadata matched screenouts data.
    """
    return match_found(
        visitor=meta,
        tracked=current_app.screenouts,
        keys=current_app.settings['screenout_keys']
    )

def is_duplicate(meta):
    """
    Returns
    -------
    match_found : bool
        Indicates that the visitor metadata was found in the current survey
        metadata.
    """
    if not current_app.settings['duplicate_keys']:
        return False
    return match_found(
        visitor=meta,
        tracked=DataStore.query.first().meta,
        keys=current_app.settings['duplicate_keys'],
    )

def match_found(visitor, tracked, keys):
    """
    Parameters
    ----------
    visitor : dict
        Visitor metadata.

    tracked : dict
        Tracked metadata from screenouts or current survey metadata.

    keys : iterable
        Keys on which to look for a match.

    Returns
    -------
    match_found : bool
        Indicates that visitor metadata matches tracked metadata on at least
        one key.
    """
    keys = keys or tracked.keys()
    for key in keys:
        visitor_val = visitor.get(key)
        tracked_vals = tracked.get(key)
        if (
            visitor_val is not None and tracked_vals is not None 
            and visitor_val in tracked_vals
        ):
            return True
    return False

@bp.route('/screenout')
def screenout():
    return current_app.settings['screenout_page']
    
@bp.route('/restart', methods=['GET','POST'])
@login_required
def restart():
    """
    Participants may navigate back to return to the in progress experiment,
    or forward to restart.
    """
    if request.method == 'POST':
        if request.form.get('direction') == 'forward':
            initialize_participant(get_metadata())
        return redirect(url_for('hemlock.'+bp.default_route))
    # request method is GET
    return current_app.settings.get('restart_page')


"""Private database models and classes

These should rarely be modified by the user.
"""

from .data_frame import DataFrame
from .data_store import DataStore
from .html_types import CSSListType, JSListType
from .router import Router
from .viewing_page import ViewingPage


"""Utilities for researcher routes"""
from ...tools import Navbar, Navitem, Navitemdd, Dropdownitem

navbar = Navbar(
    '''
    <img src="/hemlock/static/img/hemlock_favicon.svg" class="d-inline-block align-top" alt="" style="max-height:30px;">
    <span style="font-family:'Josefin Sans';">HEMLOCK</span>
    ''',
    href='https://dsbowen.github.io/hemlock',
    a_attrs={'target': '_blank'},
    navitems=[
        ('Participant status', '/status'),
        ('Download', '/download'),
        ('Logout', '/logout'),
    ]
)


"""## Default application settings

Below are the default settings for Hemlock applications and extensions.

App settings
------------
before_app_first_request : list, default=[]
    List of callables to run as the app is initialized.

clean_data : callable or None, default=None
    Callable which cleans your data before downloading or creating a data 
    profile. This callable takes and returns `pandas.DataFrame`. If `None`, no
    additional cleaning is performend.

collect_IP : bool, default=True
    Indicates that participants' IP addresses will be associated with their
    data.

duplicate_keys : list, default=[]
    List of keys (column names) on which to block duplicate participants. If
    empty, the app will not screen out duplicates.

error_500_page : str or callable, default=_gen_500_page
    Internal server error page HTML (str) or function which returns the error 
    page (str).

loading_page : str or callable, default=_gen_loading_page
    Loading page HTML (str) or function which returns the loading page (str).

restart_option : bool, default=True
    Indicates that participants who attempt to re-navigate to the index page 
    will be given the option to restart the survey. If `False`, participants 
    to attempt to re-navigate to the index page will be redirected to their 
    current survey page.

restart_page : str or callable, default=_gen_restart_page
    Restart page HTML (str) or function which returns the restart page (str).

screenout_csv : str, default='screenout.csv'
    Name of the csv file containing criteria for screening out participants.

screenout_keys : list, default=[]
    List of keys (column names) on which to screen out participants. If empty,
    participants will be screened out based on all keys in the screenout csv.

screenout_page : str or callable, default=_gen_screenout_page
    Screenout page HTML (str) or callable which returns the screen outpage 
    (str).

socket_js_src : str, default='https://cdnjs.cloudflare.com/ajax/libs/socket.io/2.3.0/socket.io.js'
    Source of the websocket javascript.

static_folder : str, default='static'
    Path to the static folder.

template_folder : str, default='templates'
    Path to the template folder.

time_expired_text : str, default='You have exceeded your time limit...'
    Text displayed to participants whose time has expired.

time_limit : datetime.timedelta or None, default=None
    Time limit for participants to complete the survey.

validate : bool, default=True
    Indicate that all validation is active. Set to `False` to turn off all 
    validation for testing.

Config
------
PASSWORD : str
    Looks for a `PASSWORD` environment variable.

PASSWORD_HASH : str
    Generated password hash.

SECRET_KEY : str
    Looks for a `SECRET_KEY` environment variable.

SQLALCHEMY_DATABASE_URI : str
    Looks for `DATABASE_URL` environment variable. Otherwise, we use a SQLite 
    database `data.db` in the current working directory.

SQLALCHEMY_TRACK_MODIFICATIONS : bool, default=False

SQLALCHEMY_ENGINE_OPTIONS : dict
    Default is empty dictionary if running on a sqlite database or 
    `{pool_size=1, pool_recycle=10, max_overflow=0}` otherwise.

REDIS_URL : str, default=None
    Looks for a `REDIS_URL` environment variable.

DownloadBtnManager
------------------

Manager
-------
loading_img_src : str, default='/hemlock/static/img/worker_loading.gif'
    Name of the loading image file.

Notes
-----
See <https://flask.palletsprojects.com/en/1.1.x/config/> for more detail on 
Flask application configuration.

See <https://dsbowen.github.io/flask-download-btn/manager/> for more detail on 
DownloadBtnManager settings.

See <https://dsbowen.github.io/flask-worker/manager/> for more detail on 
Manager settings.
"""

from flask import render_template
from werkzeug.security import generate_password_hash

import os

PASSWORD = os.environ.get('PASSWORD', '')
PASSWORD_HASH = generate_password_hash(PASSWORD)

SOCKET_JS_SRC = 'https://cdnjs.cloudflare.com/ajax/libs/socket.io/2.3.0/socket.io.js'

SQLALCHEMY_DATABASE_URI = os.environ.get(
    'DATABASE_URL', 'sqlite:///'+os.path.join(os.getcwd(), 'data.db')
)

TIME_EXPIRED_TXT = 'You have exceeded your time limit for this survey'

def _gen_500_page():
    from ..models import Page
    from ..qpolymorphs import Label

    return Page(
        Label(
            """
            <p>The application encountered an error. <b>Try refreshing the page</b> (this usually solves the problem).</p> 
            If you keep getting an error, contact the survey administrator. We apologize for the inconvenience and thank your for your patience while we resolve the issue.
            """
        ),
        forward=False
    )._render()

def _gen_loading_page():
    from ..models import Page
    from ..qpolymorphs import Label

    return Page(
        Label(
            '''
            The next page should load automatically in a few seconds.
            '''
        ),
        forward=False,
        extra_js=render_template('hemlock/loading.js')
    )._render()

def _gen_restart_page():
    from ..models import Page
    from ..qpolymorphs import Label

    return Page(
        Label(
            """
            <p>Click << to resume the survey. Click >> to restart 
            the survey.</p>
            If you restart the survey, your responses will not be 
            saved.
            """
        ),
        back=True
    )._render()

def _gen_screenout_page():
    from ..models import Page
    from ..qpolymorphs import Label

    return Page(
        Label(
            """
            <p>Our records indicate that you have already participated in this
            or similar surveys.</p>
            
            Thank you for your continuing interest in our research.
            """
        ),
        forward=False
    )._render()

settings = dict(
    before_app_first_request=[],
    clean_data=None,
    collect_IP=True,
    duplicate_keys=[],
    error_500_page=_gen_500_page,
    loading_page=_gen_loading_page,
    restart_option=True,
    restart_page=_gen_restart_page,
    screenout_csv='screenout.csv',
    screenout_keys=[],
    screenout_page=_gen_screenout_page,
    socket_js_src=SOCKET_JS_SRC,
    static_folder='static',
    template_folder=os.path.join(os.getcwd(), 'templates'),
    time_expired_text=TIME_EXPIRED_TXT,
    time_limit=None,
    validate=True,
    Config=dict(
        PASSWORD=PASSWORD,
        PASSWORD_HASH=PASSWORD_HASH,
        SECRET_KEY=os.environ.get('SECRET_KEY', 'secret-key'),
        SQLALCHEMY_DATABASE_URI=SQLALCHEMY_DATABASE_URI,
        SQLALCHEMY_TRACK_MODIFICATIONS=False,
        SQLALCHEMY_ENGINE_OPTIONS=(
            {} if SQLALCHEMY_DATABASE_URI.startswith('sqlite')
            else dict(
                pool_size=1,
                pool_recycle=10,
                max_overflow=0
            )
        ),
        REDIS_URL=os.environ.get('REDIS_URL'),
    ),
    DownloadBtnManager={},
    Manager=dict(loading_img_src='/hemlock/static/img/worker_loading.gif'),
    # Talisman=dict(
    #     content_security_policy={
    #         'default-src': ['\'self\'', '\'unsafe-inline\'', 'data:'],
    #         'frame-src': [
    #             '\'self\'', 
    #             '\'unsafe-inline\'', 
    #             'data:',
    #             'https://www.youtube.com',
    #         ],
    #         'frame-ancestors': [
    #             '*'
    #         ],
    #         'font-src': [
    #             'https://fonts.googleapis.com',
    #             'https://fonts.gstatic.com',
    #         ],
    #         'img-src': ['\'self\'', '\'unsafe-inline\'', 'data:'],
    #         'script-src': [
    #             '\'self\'',
    #             '\'unsafe-eval\'',
    #             '\'unsafe-inline\'',
    #             'https://code.jquery.com',
    #             'https://cdn.jsdelivr.net',
    #             'https://stackpath.bootstrapcdn.com',
    #             'https://cdnjs.cloudflare.com'
    #         ],
    #         'style-src': [
    #             '\'self\'',
    #             '\'unsafe-inline\'',
    #             'https://fonts.googleapis.com/css',
    #             'https://stackpath.bootstrapcdn.com'
    #         ],
    #     },
    # )
)


"""# File upload"""

from ..app import db, settings
from ..models import Question
from ..tools import join
from .bases import InputBase

from flask import current_app, render_template, request
from sqlalchemy_mutable import HTMLAttrsType, MutableListJSONType
from werkzeug.utils import secure_filename

import os
from datetime import timedelta
from mimetypes import types_map

INVALID_MIMETYPE_MSG = """
<p>Please upload a file with one of the following extensions: {}.
"""

def upload_to_bucket(file_):
    """
    Default `hemlock.File` submit function. Uploads a participant file to 
    Google bucket.

    Parameters
    ----------
    file_ : hemlock.File
    """
    path = file_.get_path()
    upload = request.files.get(file_.model_id)
    if not (path and upload):
        return
    blob = current_app.gcp_bucket.blob(path)
    if path.endswith('.txt'):
        # .txt files require mimetype 'text/html' 
        # rather than the default 'text/plain'
        blob.upload_from_string(upload.read(), content_type='text/html')
        return
    blob.upload_from_string(upload.read())  
        
settings['File'] = dict(
    input_attrs={'class': ['custom-file-input'], 'type': 'file'},
    file_label_attrs={'class': ['label-txt', 'custom-file-label']},
    submit_functions=upload_to_bucket
)


class File(InputBase, Question):
    """
    Allows participants to upload files.

    Inherits from [`hemlock.qpolymorphs.InputBase`](bases.md)
    and [`hemlock.Question`](../models/question.md).

    Parameters
    ----------
    label : str or None, default=None
        Upload file label.

    template : str, default='hemlock/file.html'
        Template for the file upload body.

    Attributes
    ----------
    allowed_extensions : list
        Allowed file extensions, e.g. `['.png','.jpeg']`.

    filename : str
        Name of the file as stored in the Google bucket.

    Examples
    --------
    Set up a 
    [Google bucket](https://cloud.google.com/storage/docs/creating-buckets)
    with the appropriate 
    [CORS permissions](https://cloud.google.com/storage/docs/cross-origin).

    Set an environment variable `BUCKET` to the name of the bucket, and 
    `GOOGLE_APPLICATION_CREDENTIALS` to the name of your 
    [Google application credentials JSON file](https://cloud.google.com/docs/authentication/getting-started).

    ```
    export BUCKET=<my-bucket> GOOGLE_APPLICATION_CREDENTIALS=<my-credentials.json>
    ```

    In `survey.py`:

    ```python
    from hemlock import Branch, File, Page, Label, route

    @route('/survey')
    def start():
    \    return Branch(
    \        Page(File(
    \            '<p>Upload a .png</p>', 
    \            filename='upload',
    \            allowed_extensions=['.png']
    \        )),
    \        Page(Label('<p>The End</p>'), terminal=True)
    \    )
    ```
    
    In `app.py`:

    ```python
    import survey

    from hemlock import create_app

    app = create_app()

    if __name__ == '__main__':
    \    from hemlock.app import socketio
    \    socketio.run(app, debug=True)
    ```

    Run the app locally with:

    ```
    $ python app.py # or python3 app.py
    ```

    And open your browser to <http://localhost:5000/>. Upload a .png and click to the next page. You'll find your uploaded file in your Google bucket in `participant-1/upload.png`.
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'file'}

    # HTML attributes
    file_label = db.Column(db.String)
    file_label_attrs = db.Column(HTMLAttrsType)

    # Additional attributes
    allowed_extensions = db.Column(MutableListJSONType)
    filename = db.Column(db.String)

    def __init__(self, label=None, template='hemlock/file.html', **kwargs):
        super().__init__(label=label, template=template, **kwargs)
        self.js.append(render_template('hemlock/file.js', q=self))

    def generate_signed_url(self, expiration=timedelta(hours=.5), **kwargs):
        """
        Generate a signed URL for the uploaded file.

        Parameters
        ----------
        expiration : datetime.timedelta, default=datetime.timedelta(0, 1800)
            Duration for which this signed URL is valid.

        \*\*kwargs :
            Additional keyword arguments are passed to the 
            `generate_signed_url` method.
        
        Notes
        -----
        Read more about [signed URLs](https://cloud.google.com/storage/docs/access-control/signed-urls).
        """
        path = self.get_path()
        if not (path and self.response):
            return
        blob = current_app.gcp_bucket.blob(path)
        return blob.generate_signed_url(expiration=expiration, **kwargs)

    def get_allowed_types(self):
        """
        Returns
        -------
        allowed_types : list
            List of allowed mimetypes. Derived from `self.allowed_extensions`.
        """
        if self.allowed_extensions:
            return set([types_map[ext] for ext in self.allowed_extensions])

    def get_path(self):
        """        
        Returns
        -------
        path : str
            Path to the uploaded file in the Google bucket.
        """
        if not self.filename:
            return
        if self.filename.startswith('/'):
            # absolute path
            path = self.filename
        else:
            # relative path in participant's folder
            # participant's folder is `part.model_id`
            path = os.path.join(self.part.model_id, self.filename)
        if self.response and '.' in self.response:
            # add the extension of the participant's response to the path
            # note: response is the name of the uploaded file
            path += '.' + self.response.split('.')[-1]
        return path

    def _record_response(self):
        """Record response as name of uploaded file"""
        file_ = request.files.get(self.model_id)
        if file_ is None:
            # no file uploaded
            self.response = None
            return
        self.response = secure_filename(file_.filename)

    def _validate(self):
        """Verify that the uploaded file's mimetype is allowed"""
        if self.response and self.allowed_extensions:
            file_ = request.files.get(self.model_id)
            if file_.mimetype not in self.get_allowed_types():
                self.error = INVALID_MIMETYPE_MSG.format(
                    join('or', *self.allowed_extensions)
                )
                return False
        return super()._validate()


from sqlalchemy.types import JSON
from sqlalchemy_mutable import MutableList


class CSSListType(JSON):
    pass


class CSSList(MutableList):
    """
    Items in a `CSSList` must be one of the following:

    1. Link tag (str) e.g., `'<link rel="stylesheet" href="https://my-css-url">'`
    2. Href (str) e.g., `"https://my-css-url"`
    3. Style dictionary (dict) e.g., `{'body': {'background': 'coral'}}`

    The style dictionary maps a css selector to an attributes dictionary. The attributes dictionary maps attribute names to values.
    """
    def render(self):
        def render_item(item):
            if isinstance(item, str):
                # interpret item as a link tag or href
                return (
                    item if item.startswith('<link')
                    else '<link rel="stylesheet" href="{}">'.format(item)
                )

            # interpret item as style dictionary
            def format_style(selector, attrs):
                attrs = ' '.join(
                    ['{}: {};'.format(*item) for item in attrs.items()]
                )
                return selector+' {'+attrs+'}'

            css = ' '.join(
                [format_style(key, val) for key, val in item.items()]
            )
            return '<style>{}</style>'.format(css)

        return '\n'.join([render_item(item) for item in self])


CSSList.associate_with(CSSListType)


class JSListType(JSON):
    pass


class JSList(MutableList):
    """
    Items in a `JSList` must be one of the following:

    1. Attributes dictionary (dict) e.g., `{'src': 'https://my-js-url'}`
    2. JS code (str)
    3. Tuple of (attributes dictionary, js code)
    """
    def render(self):
        from ...tools import format_attrs

        def render_item(item):
            assert isinstance(item, (tuple, dict, str))
            if isinstance(item, tuple):
                attrs, js = item
            elif isinstance(item, dict):
                attrs, js = item, ''
            elif isinstance(item, str):
                if item.startswith('<script'):
                    # item is already formatted
                    return item
                attrs, js = {}, item
            return '<script {}>{}</script>'.format(format_attrs(**attrs), js)

        return '\n'.join([render_item(item) for item in self])


JSList.associate_with(JSListType)


"""# Application factory and settings"""

from .settings import settings

import pandas as pd
from flask import Flask, Blueprint, current_app
# from flask_apscheduler import APScheduler
from flask_download_btn import DownloadBtnManager
from flask_login import LoginManager
from flask_socketio import SocketIO
from flask_sqlalchemy import SQLAlchemy
from flask_worker import Manager
from sqlalchemy_mutable import MutableManager

import os

bp = Blueprint(
    'hemlock', 
    __name__,
    template_folder='templates',
    static_folder='static',
    static_url_path='/hemlock/static'
)
db = SQLAlchemy()
download_btn_manager = DownloadBtnManager(db=db)
bucket = os.environ.get('BUCKET')
if bucket is not None:
    from google.cloud import storage
    gcp_client = storage.Client()
    gcp_bucket = gcp_client.get_bucket(bucket)
login_manager = LoginManager()
login_manager.login_view = 'hemlock.index'
login_manager.login_message = None
# scheduler = APScheduler()
socketio = SocketIO(async_mode='eventlet')
manager = Manager(db=db, socketio=socketio)
MutableManager.db = db

@bp.before_app_first_request
def init_app():
    """
    Create database tables and initialize data storage models
    """
    from ..models.private import DataStore

    db.create_all()
    # cache static pages
    static_pages = [
        'error_500_page', 'loading_page', 'restart_page', 'screenout_page'
    ]
    for page_key in static_pages:
        page = current_app.settings[page_key]
        if callable(page):
            current_app.settings[page_key] = page()
    # create datastore
    if not DataStore.query.first():
        db.session.add(DataStore())
    # run additional before app first request functions
    [f() for f in current_app.settings['before_app_first_request']]
    db.session.commit()

def push_app_context():
    """
    Push an app context for debugging in shell or notebook.

    Returns
    -------
    app : flask.app.Flask

    Examples
    --------
    ```python
    from hemlock import push_app_context

    app = push_app_context()
    ```

    Out:

    ```
    <Flask 'hemlock.app'>
    ```
    """    
    app = create_app()
    app.app_context().push()
    app.test_request_context().push()
    init_app()
    return app

def create_app(settings=settings):
    """
    Create a Hemlock application.

    Parameters
    ----------
    settings : dict
        Default settings for the application, extensions, and models.

    Returns
    -------
    app : flask.app.Flask

    Examples
    --------
    In this example, we add a back button to every page in our survey.

    ```python
    from hemlock import Page, create_app, settings

    settings['Page'].update({'back': True})

    app = create_app()
    app.app_context().push()
    Page().preview()
    ```
    """
    app = _create_app(settings)
    _set_bucket(app)
    _set_redis(app)
    _init_extensions(app, settings)
    app.tmpfiles = []   
    return app

def _create_app(settings):
    app = Flask(
        __name__, 
        static_folder=settings.get('static_folder'), 
        template_folder=settings.get('template_folder'),
    )
    # get screenouts
    screenout_csv = settings.get('screenout_csv')
    if os.path.exists(screenout_csv):
        df = pd.read_csv(screenout_csv)
        screenout_keys = settings.get('screenout_keys')
        df = df[screenout_keys] if screenout_keys else df
        app.screenouts = df.to_dict(orient='list')
    else:
        app.screenouts = {}
    # store configuration, settings, and blueprint
    app.config.update(settings.get('Config'))
    app.settings = settings
    app.register_blueprint(bp)
    return app

def _set_bucket(app):
    """Set up google bucket"""
    if bucket is not None:
        app.gcp_client, app.gcp_bucket = gcp_client, gcp_bucket
    else:
        app.gcp_client = app.gcp_bucket = None

def _set_redis(app):
    """Set up Redis Queue"""
    redis_url = app.config.get('REDIS_URL')
    if redis_url is not None:
        from redis import ConnectionPool, Redis 
        from rq import Queue

        connection_pool = ConnectionPool.from_url(redis_url)
        app.redis = Redis(connection_pool=connection_pool, ssl_cert_reqs=None)
        app.task_queue = Queue('hemlock-task-queue', connection=app.redis)
    else:
        app.redis = app.task_queue = None

def _init_extensions(app, settings):
    """Initialize application extensions"""
    db.init_app(app)
    download_btn_manager.init_app(app, **settings.get('DownloadBtnManager'))
    login_manager.init_app(app)
    # scheduler.init_app(app)
    # scheduler.start()
    socketio.init_app(app, message_queue=app.config.get('REDIS_URL'))
    manager.init_app(app, **settings.get('Manager'))


"""# Blank"""

from .bases import InputBase
from ..app import db, settings
from ..functions.debug import send_keys
from ..models import Question
from ..tools import key

from flask import render_template

settings['Blank'] = {
    'class': ['form-control'], 'type': 'text', 'blank_empty': '', 'debug': send_keys
}


class Blank(InputBase, Question):
    """
    Fill in the blank question.

    Parameters
    ----------
    label : tuple, list, or None, default=None
        If the label is a `tuple` or `list`, the participant's response will 
        fill in the blanks between items.

    template : str, default='hemlock/input.html'
        File name of the Jinja template.

    blank_empty : str, default=''
        String used to fill in the blank when the participant's response is 
        empty.

    Examples
    --------
    ```python
    from hemlock import Page, Blank

    Page(
    \    Blank(
    \        ('Hello, ', '!'),
    \        blank_empty='_____'
    \    )
    ).preview()
    ```

    Enter 'World' in the input and the label will change to 'Hello, World!'.
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'blank'}

    blank_id = db.String(8)
    blank_empty = db.String()

    def __init__(self, label=None, template='hemlock/input.html', **kwargs):
        self.blank_id = key(8)
        super().__init__(label=label, template=template, **kwargs)
        self.js.append(render_template('hemlock/blank.js', q=self))

    def __setattr__(self, key, val):
        if key == 'label' and isinstance(val, (tuple, list)):
            val = '<span name="{}"></span>'.format(self.blank_id).join(val)
        super().__setattr__(key, val)


"""# Select (dropdown)"""

from ..app import db, settings
from ..functions.debug import click_choices
from ..models import ChoiceQuestion
from .bases import InputBase
from .choice import OptionListType

from sqlalchemy_mutable import HTMLAttrsType

settings['Select'] = dict(
    input_attrs={'class': ['custom-select'], 'multiple': False},
    debug=click_choices,
)


class Select(InputBase, ChoiceQuestion):
    """
    Select questions allow participants to select one or more options from a 
    dropdown menu.

    Inherits from [`hemlock.InputBase`](bases.md) and 
    [`hemlock.ChoiceQuestion`](../models/question.md).

    Its default debug function is
    [`click_choices`](../functions/debug.md#hemlockfunctionsdebugclick_choices).

    Parameters
    ----------
    label : str or None, default=None
        Select question label.

    choices : list of hemlock.Option, str, tuple, or dict, default=[]
        Options the participant can select.

    template : str, default='hemlock/select.html'
        Template for the select body.

    Attributes
    ----------
    align : str, default='left'
        Choice alignment; `'left'`, `'center'`, or `'right'`.

    choices : list of `hemlock.Option`
        Set for the `choices` parameter.
        
    multiple : bool, default=False
        Indicates that the participant may select multiple choices.

    Examples
    --------
    ```python
    from hemlock import Page, Select, Option, push_app_context

    app = push_app_context()

    Page(Select('<p>Select one.</p>', ['World','Moon','Star'])).preview()
    ```

    Notes
    -----
    An option can be input as any of the following:

    1. `Option` object.
    2. `str`, treated as the choice label.
    3. `(choice label, value)` tuple.
    4. `(choice label, value, name)` tuple.
    5. Dictionary with choice keyword arguments.
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'select'}
    options = db.Column(OptionListType)

    @property
    def choices(self):
        return self.options

    @choices.setter
    def choices(self, val):
        self.options = val

    _input_attr_names = [
        'autofocus',
        'disabled',
        'multiple',
        'required',
        'size'
    ]

    def __init__(
            self, label=None, choices=[], template='hemlock/select.html', 
            **kwargs
        ):
        super().__init__(label, choices, template, **kwargs)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hemlock-survey",
    version="0.0.45",
    author="Dillon Bowen",
    author_email="dsbowen@wharton.upenn.edu",
    description="A package for creating and deploying surveys",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://dsbowen.github.io/hemlock",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        'bs4>=0.0.1',
        'convert-list>=0.0.1',
        'cssutils>=1.0.2',
        'eventlet>=0.25.1',
        'flask>=1.1.1',
        'flask-download-btn>=0.0.22',
        'flask-login>=0.4.1',
        'flask-socketio==4.3.2', # 5.0.0 and later is giving a strange error message; check back for resolution https://github.com/miguelgrinberg/Flask-SocketIO/issues/1432
        'flask-sqlalchemy>=2.4.1',
        'flask-talisman>=0.7.0',
        'flask-worker>=0.0.14',
        'pandas>=1.0.1',
        'python-docx>=0.8.10',
        'redis>=3.3.11',
        'rq>=1.2.0',
        'selenium>=3.141.0',
        'selenium-tools>=0.0.1',
        'sqlalchemy>=1.3.12',
        'sqlalchemy-modelid>=0.0.3',
        'sqlalchemy-mutable>=0.0.11',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


"""# Embedded data and timers"""

from ..app import db
from .bases import Data

from datetime import datetime, timedelta


class Embedded(Data):
    """
    Embedded data belong to a branch or page. Use embedded data to manually 
    input data to the dataframe; as opposed to recording data from participant 
    responses.

    Polymorphic with [`hemlock.models.Data`](bases.md).

    Parameters
    ----------
    var : str or None, default=None
        Variable name associated with this data element. If `None`, the data 
        will not be recorded.

    data : sqlalchemy_mutable.MutableType
        Data this element contributes to the dataframe.

    data_rows : int, default=1
        Number of rows this data element contributes to the dataframe for its 
        participant. If negative, this data element will 'fill in' any emtpy
        rows at the end of the dataframe with its most recent value.

    Relationships
    -------------
    participant : hemlock.Participant or None
        The participant to whom this data element belongs.

    branch : hemlock.Branch or None
        The branch to which the embedded data element belongs.

    page : hemlock.Page or None
        The page to which this embedded data element belongs.

    Examples
    --------
    ```python
    from hemlock import Branch, Embedded, Page, Participant, push_app_context

    def start():
    \    return Branch(Page())

    app = push_app_context()

    part = Participant.gen_test_participant(start)
    part.embedded = [Embedded('Name', 'Socrates', data_rows=-1)]
    part.get_data()
    ```

    Out:

    ```
    {'ID': [1],
    'EndTime': [datetime.datetime(2020, 7, 4, 17, 57, 23, 854272)],
    'StartTime': [datetime.datetime(2020, 7, 4, 17, 57, 23, 854272)],
    'Status': ['InProgress'],
    'Name': ['Socrates'],
    'NameOrder': [0],
    'NameIndex': [0]}
    ```
    """
    id = db.Column(db.Integer, db.ForeignKey('data.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'embedded'}

    _part_id = db.Column(db.Integer, db.ForeignKey('participant.id'))
    _page_id = db.Column(db.Integer, db.ForeignKey('page.id'))
    _branch_id = db.Column(db.Integer, db.ForeignKey('branch.id'))

    def __init__(self, var=None, data=None, data_rows=1, **kwargs):
        self.var = var
        self.data = data
        self.data_rows = data_rows
        super().__init__(**kwargs)


class Timer(Embedded):
    """
    Tracks how much time a participant spends in various parts of the survey.

    Inherits from `hemlock.Embedded`.

    Attributes
    ----------
    data : float or None
        Read only. Number of seconds for which the timer has been running.

    end_time : datetime.datetime or None
        Read only. If the timer is running, this is the current time. If the 
        timer is paused, this is the time at which the timer was last paused.

    start_time : datetime.datetime or None
        The time at which the timer was started.

    state : str
        `'not started'`, `'running`', or `'paused'`.

    total_time : datetime.timedelta or None
        Read only. Total time the timer has been running.

    Examples
    --------
    ```python
    from hemlock import Timer, push_app_context

    import time

    app = push_app_context()

    timer = Timer()
    print(timer.state)
    timer.start()
    print(timer.state)
    time.sleep(1)
    print(timer.data)
    timer.pause()
    print(timer.state)
    time.sleep(1)
    print(timer.data)
    ```

    Out:

    ```
    not started
    running
    1.002405
    paused
    1.002983
    ```
    """
    id = db.Column(db.Integer, db.ForeignKey('embedded.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'timer'}

    _timed_page_id = db.Column(db.Integer, db.ForeignKey('page.id'))

    _data = db.Column(db.Float)
    _end_time = db.Column(db.DateTime)
    start_time = db.Column(db.DateTime)
    state = db.Column(db.String(16), default='not started')
    _total_time = db.Column(db.Interval)

    @property
    def data(self):
        self._set_current_time()
        return self._data

    @property
    def end_time(self):
        self._set_current_time()
        return self._end_time

    @property
    def total_time(self):
        self._set_current_time()
        return self._total_time

    def __init__(self, var=None, data_rows=1, **kwargs):
        self.var = var
        self.data_rows = data_rows
        [setattr(self, key, val) for key, val in kwargs]

    def start(self):
        """
        Start the timer.

        Returns
        -------
        self : hemlock.Timer
        """
        if self.state != 'running':
            self.state = 'running'
            now = datetime.utcnow()
            if self.start_time is None:
                self.start_time = now
            self._end_time = now
        return self
    
    def pause(self):
        """
        Pause the timer.

        Returns
        -------
        self : hemlock.Timer
        """
        self._set_current_time()
        self.state = 'paused'
        return self

    def reset(self):
        """
        Reset all attributes to their default values.

        Returns
        -------
        self : hemlock.Timer
        """
        self._data = None
        self._end_time = None
        self.state = 'not started'
        self.start_time = None
        self._total_time = None
        return self
    
    def _set_current_time(self):
        if self.state == 'running':
            now = datetime.utcnow()
            if self._total_time is None:
                self._total_time = timedelta(0)
            self._total_time += now - self._end_time
            self._end_time = now
            self._data = self._total_time.total_seconds()


"""# Application setup"""

from ..app import bp, db
from ..models import Participant
from . import participant
from . import researcher

from flask import current_app, session, request
from flask_login import current_user, login_required

from functools import wraps

@bp.errorhandler(500)
def internal_server_error(e):
    # set in_progress to False to let the router know it should retry when the
    # participant refreshes the page
    current_user._router.in_progress = False
    db.session.commit()
    return current_app.settings['error_500_page'], 500

def route(path, default=False):
    """Decorator for registering a view function

    This decorator wraps a navigate function and converts it into a view 
    function.
    """
    def register_view(gen_root):
        if default or not hasattr(bp, 'default_route'):
            bp.default_route = gen_root.__name__
            
        @bp.route(path, methods=['GET','POST'])
        @login_required
        @wraps(gen_root)
        def view():
            part = get_part()
            if not part.branch_stack:
                part._init_tree(gen_root)
            return part._router()

        return gen_root

    return register_view

def get_part():
    """Get participant

    Normally, a participant is the `current_user` from Flask-Login.

    However, for testing, it is often useful to have multiple surveys open 
    simultaneously. To support this, a participant can also be retrieved by 
    `id` and `_key`. Navigate to {URL}/?Test=1.
    """
    if request.method == 'GET':
        part_id = request.args.get('part_id')
        part_key = request.args.get('part_key')
    else: # request method is POST
        part_id = request.form.get('part_id')
        part_key = request.form.get('part_key')
    if part_id is None:
        part = current_user
    else:
        part = Participant.query.get(part_id)
        assert part_key == part._key
    return part


"""# MTurk utilities"""

from .random import key

import numpy as np
import pandas as pd
from flask_login import current_user

ID_LABEL = '<p>Enter your MTurk ID to consent</p>'
CONFIRM_LABEL = '<p>Confirm your ID</p>'

def consent_page(
        consent_label=None, 
        id_label=ID_LABEL, 
        confirm_label=CONFIRM_LABEL, 
        require=True
    ):
    """
    Parameters
    ----------
    consent_label : str or None, default=None
        The consent terms for your study. If `None`, the consent label is
        omitted.

    id_label : str, default='Enter your MTurk ID to consent'
        Label asking workers to enter their ID.

    confirm_label : str, default='Confirm your ID'
        Label asking for confirmation.

    require : bool, default=True
        Indicates that workers are required to enter their ID.

    Returns
    -------
    consent page : hemlock.Page
    """
    from ..models import Page, Debug as D, Validate as V
    from ..qpolymorphs import Label, Input

    id_q = Input(
        id_label, 
        required=require, 
        debug=D.send_keys('test'), 
        submit=_record_id
    )
    page = Page(
        id_q,
        Input(
            confirm_label, 
            required=require,
            debug=D.send_keys('test'),
            validate=V.match(
                id_q, error_msg='IDs do not match'
            )
        )
    )
    if consent_label:
        page.questions.insert(0, Label(consent_label))
    return page

def _record_id(id_input):
    current_user.meta['WorkerId'] = id_input.data.strip()

def completion_page(participant=None):
    """
    Parameters
    ----------
    participant : hemlock.Participant or None, default=None
        Record the completion code in this participant's metadata. If `None`,
        the completion code is recorded in flask-login's `current_user`.

    Returns
    -------
    completion page: hemlock.Page
        Completion page with unique completion code.
    """
    from ..models import Page
    from ..qpolymorphs import Label

    code = key(6)
    part = participant or current_user
    part.meta['SurveyCode'] = code
    return Page(
        Label(
            '''
            <p>Thank you for completing the study. Your completion code is 
            <b>{}</b></p>
            
            The completion code is case-sensitive.
            '''.format(code)
        ),
        terminal=True
    )

def get_approve_df(data_df, batch_df, bonus=False, verbose=True):
    """
    Parameters
    ----------
    data_df : str or pd.DataFrame
        Hemlock survey data. This must have `'WorkerId'` and `'SurveyCode'` 
        columns.

    bonus : bool, default=False
        Indicates that workers will receive a bonus. If bonusing workers, `data_df`` must include a column called `'BonusAmount'`.

    verbose : bool, default=True
        Indicates to print information on approvals, rejections, and bonuses.

    Returns
    -------
    approve_df : pd.DataFrame
        Dataframe with approval and bonus information.

    Examples
    --------
    ```python
    from hemlock.tools import get_approve_df, approve_assignments

    approve_df = get_approve_df(data_df, 'path/to/batch.csv', bonus=True)
    ```

    Out:

    ```
    90 assignments approved
    10 assignments rejected
    Total bonus: $90
    ```

    ```python
    import boto3

    client = boto3.client('mturk')
    approve_assignments(client, approve_df, bonus=True, reason='Great job!')
    """
    if isinstance(data_df, str):
        data_df = pd.read_csv(data_df)
    data_df = data_df.drop_duplicates('WorkerId')
    if isinstance(batch_df, str):
        batch_df = pd.read_csv(batch_df)
    data_cols = ['WorkerId', 'SurveyCode', 'Status', 'ID']
    if bonus:
        data_cols.append('BonusAmount')
    df = batch_df.merge(
        data_df[data_cols], how='left', on='WorkerId'
    )
    df['Approve'] = df.SurveyCode == df['Answer.surveycode']
    if bonus:
        df['UniqueRequestToken'] = np.random.rand(len(df))
    if verbose:
        print('{} assignments approved'.format(df.Approve.sum()))
        print('{} assignments rejected'.format((~df.Approve).sum()))
        if bonus:
            print('Total bonus: ${:.2f}'.format(df.BonusAmount.sum()))
    return df

def approve_assignments(
        client, approve_df, approve=True, bonus=False, bonus_reason='', 
        OverrideRejection=False
    ):
    """
    Approve and reject assignments and pay bonuses.

    Parameters
    ----------
    client : boto3.client

    approve_df : pd.DataFrame
        Output of `get_approve_df`.

    approve : bool, default=True
        Indicates that assignments should be approved and rejected.

    bonus : bool, default=False
        Indicates that bonuses should be paid.

    bonus_reason : str, default=''
        Reason for giving bonuses. This must be nonempty if paying bonuses.

    OverrideRejection : bool, default=False
        Indicates that this function can override rejected assignments.
    """
    assert not bonus or bonus_reason, 'Please provide a reason for bonusing workers'
    def approve_assignment(x):
        if x.Approve:
            client.approve_assignment(
                AssignmentId=x.AssignmentId,
                OverrideRejection=OverrideRejection
            )
        else:
            client.reject_assignment(
            AssignmentId=x.AssignmentId,
            RequesterFeedback='''
                Could not verify completion code. If you believe this is a 
                mistake, please contact the requester.
                '''.strip()
        )
        
    def send_bonus(x):
        if x.Approve and x.BonusAmount > 0:
            client.send_bonus(
                WorkerId=x.WorkerId,
                BonusAmount=str(round(x.BonusAmount, 2)),
                AssignmentId=x.AssignmentId,
                Reason=bonus_reason,
                UniqueRequestToken=str(x.UniqueRequestToken)
            )
        
    if approve:
        approve_df.apply(approve_assignment, axis=1)
    if bonus:
        approve_df.apply(send_bonus, axis=1)


"""# Language tools"""

def indef_article(word):
    """
    Parameters
    ----------
    word : str
        Word to which the indefinite article belongs.

    Returns
    -------
    article : str
        `'an'` if `word` starts with a vowel, or `'a'` otherwise, followed 
        by the word.

    Examples
    --------
    ```python
    from hemlock.tools import indef_article

    [indef_article(fruit) for fruit in ('apple','banana')]
    ```

    Out:

    ```
    ['an apple', 'a banana']
    ```
    """
    return 'an '+word if word[0] in 'aeiou' else 'a '+word

def join(joiner, *items):
    """
    Parameters
    ----------
    joiner : str
        Joins the first n-1 items with the last item, e.g. `'and'`.

    \*items : str
        Items to join.

    Returns
    -------
    joined : str
        Joined items.

    Examples
    --------
    ```python
    from hemlock.tools import join

    print(join('and', 'world', 'sun'))
    print(join('or', 'world', 'sun', 'moon'))
    ```

    Out:

    ```
    world and sun
    world, sun, or moon
    ```
    """
    if not items:
        return ''
    if len(items) == 1:
        return str(items[0])
    if len(items) == 2:
        return '{} {} {}'.format(items[0], joiner, items[1])
    string = ', '.join([str(i) for i in items[:-1]])
    string += ', {} {}'.format(joiner, str(items[-1]))
    return string

def plural(n, singular, plural=None):
    """
    Parameters
    ----------
    n : int
        Number.

    singular : str
        The singular form of the word.

    plural : str or None, default=None
        The plural form of the word. If `None`, the plural form is assumed to
        be the singular + 's'.

    Returns
    -------
    word : str
        The singular form if number is 1, plural form otherwise.

    Examples
    --------
    ```python
    from hemlock.tools import plural

    ['{} {}'.format(n, plural(n, 'cat')) for n in range(0,3)]
    ```

    Out:

    ```
    ['0 cats', '1 cat', '2 cats']
    ```
    """
    return singular if n==1 else (plural or singular+'s')

def pronouns(person, singular, gender=None, pfx=''):
    """
    Parameters
    ----------
    person : int
        `1`, `2`, or `3` for first, second, or third person.

    singular : bool
        `True` for singular, `False` for plural.

    gender : str or None, default=None
        `'male'`, `'female'`, `'neuter'`, or `'epicene'`. Required for third 
        person singular pronouns.

    pfx : str, default=''
        Prefix for dictionary keys. Use this to distinguish between multiple
        entities in a single string.

    Returns
    -------
    pronouns : dict
        Mapping of pronoun keys to pronouns. Pronoun keys are `'subject'`, 
        `'object'`, `'dep_poss'` (dependent possessive), `'indep_poss'`,
        (indepedent possessive), `'reflex'` (reflexive).

    Examples
    --------
    ```python
    from hemlock.tools import pronouns

    string = '''
    {A_subject} said hello to {B_object} as {B_subject} was walking 
    {B_dep_poss} neighbor's dog. 'Is that dog {B_indep_poss}?', {A_subject}
    thought to {A_reflex}.
    '''
    string.format(
    \    **pronouns(3, True, 'male', pfx='A_'),
    \    **pronouns(3, True, 'female', pfx='B_')
    )
    ```

    Out:

    ```
    he said hello to her as she was walking 
    her neighbor's dog. 'Is that dog hers?', he
    thought to himself.
    ```

    In:

    ```python
    string.format(
    \    **pronouns(3, True, 'female', pfx='A_'),
    \    **pronouns(3, True, 'male', pfx='B_')
    )
    ```

    Out:

    ```
    she said hello to him as he was walking 
    his neighbor's dog. 'Is that dog his?', she
    thought to herself.
    ```
    """
    selected = pronouns_[person]['singular' if singular else 'plural']
    selected = selected[gender] if person == 3 and singular else selected
    return {str(pfx)+key: val for key, val in selected.items()}

pronouns_ = {
    1: {
        'singular': {
            'subject': 'I',
            'object': 'me',
            'dep_poss': 'my',
            'indep_poss': 'mine',
            'reflex': 'myself',
        },
        'plural': {
            'subject': 'we',
            'object': 'us',
            'dep_poss': 'our',
            'indep_poss': 'ours',
            'reflex': 'ourselves',
        }
    },
    2 : {
        'singular': {
            'subject': 'you',
            'object': 'you',
            'dep_poss': 'your',
            'indep_poss': 'yours',
            'reflex': 'yourself',
        },
        'plural': {
            'subject': 'you',
            'object': 'you',
            'dep_poss': 'your',
            'indep_poss': 'yours',
            'reflex': 'yourselves',
        },
    },
    3: {
        'singular': {
            'male': {
                'subject': 'he',
                'object': 'him',
                'dep_poss': 'his',
                'indep_poss': 'his',
                'reflex': 'himself',
            },
            'female': {
                'subject': 'she',
                'object': 'her',
                'dep_poss': 'her',
                'indep_poss': 'hers',
                'reflex': 'herself',
            },
            'neuter': {
                'subject': 'it',
                'object': 'it',
                'dep_poss': 'its',
                'indep_poss': 'its',
                'reflex': 'itself',
            },
            'epicene': {
                'subject': 'they',
                'object': 'them',
                'dep_poss': 'their',
                'indep_poss': 'theirs',
                'reflex': 'themself',
            }
        },
        'plural': {
            'subject': 'they',
            'object': 'them',
            'dep_poss': 'their',
            'indep_poss': 'theirs',
            'reflex': 'themselves',
        }
    }
}